import dataclasses
import datetime
import hashlib
import pathlib
import subprocess
import sys
from os import PathLike
from typing import Optional

from uqbar.graphs import Grapher

import supriya

from .typing import SupportsRender


@dataclasses.dataclass(frozen=True)
class PlayMemo:
    contents: bytes
    suffix: str

    def __render__(
        self,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        **kwargs,
    ) -> pathlib.Path:
        if output_file_path is None:
            hexdigest = hashlib.sha1(self.contents).hexdigest()
            file_name = f"audio-{hexdigest}{self.suffix}"
            file_path = (
                pathlib.Path(render_directory_path or supriya.output_path) / file_name
            )
        else:
            file_path = pathlib.Path(output_file_path)
        file_path.write_bytes(self.contents)
        return file_path

    @classmethod
    def from_path(cls, path: pathlib.Path) -> "PlayMemo":
        return cls(contents=path.read_bytes(), suffix=path.suffix)


class Player:

    ### INITIALIZER ###

    def __init__(self, renderable, **kwargs):
        self.renderable = renderable
        self.render_kwargs = kwargs

    ### SPECIAL METHODS ###

    def __call__(self):
        output_path = self.render()
        self.open_output_path(output_path)
        return output_path

    ### PUBLIC METHODS ###

    def open_output_path(self, output_path):
        viewer = "open -a 'QuickTime Player'"
        if sys.platform.lower().startswith("linux"):
            viewer = "xdg-open"
        subprocess.run(f"{viewer} {output_path}", shell=True, check=True)

    def render(self):
        result = self.renderable.__render__(**self.render_kwargs)
        if hasattr(result, "__render__"):
            result = result.__render__(**self.render_kwargs)
        return result


class Plotter:

    ### INITIALIZER ###

    def __init__(self, plottable, **kwargs):
        self.plottable = plottable
        self.plot_kwargs = kwargs

    ### SPECIAL METHODS ###

    def __call__(self):
        output_path = self.render()
        self.open_output_path(output_path)
        return output_path

    ### PUBLIC METHODS ###

    def open_output_path(self, output_path):
        viewer = "open"
        if sys.platform.lower().startswith("linux"):
            viewer = "xdg-open"
        subprocess.run(f"{viewer} {output_path}", shell=True, check=True)

    def render(self):
        import librosa.display
        from matplotlib import pyplot

        array, sample_rate = self.plottable.__plot__()
        fig, ax = pyplot.subplots(nrows=1)
        librosa.display.waveshow(array, sr=sample_rate, ax=ax)
        timestamp = (
            datetime.datetime.now().isoformat().replace(".", "-").replace(":", "-")
        )
        extension = self.plot_kwargs.get("format_", "png")
        output_path = supriya.output_path / f"{timestamp}.{extension}"
        fig.savefig(output_path)
        return output_path


def graph(graphable, format_="pdf", layout="dot"):
    return Grapher(
        graphable, format_=format_, layout=layout, output_directory=supriya.output_path
    )()


def play(renderable: SupportsRender, **kwargs):
    return Player(renderable, **kwargs)()


def plot(plottable, format_="png", **kwargs):
    return Plotter(plottable, format_=format_, **kwargs)()


def render(
    renderable: SupportsRender,
    output_file_path: Optional[PathLike] = None,
    render_directory_path: Optional[PathLike] = None,
    **kwargs,
):
    return renderable.__render__(
        output_file_path=output_file_path,
        render_directory_path=render_directory_path,
        **kwargs,
    )


__all__ = ["Player", "Plotter", "graph", "play", "plot", "render"]



"""
Tools for interacting with soundfiles.
"""
import aifc
import hashlib
import pathlib
import shlex
import sndhdr
import subprocess
import wave
from os import PathLike
from typing import Optional

import uqbar.strings

import supriya
from supriya.system import SupriyaObject, SupriyaValueObject


class Say(SupriyaValueObject):
    """
    Wrapper for OSX ``say`` command.

    ::

        >>> say = supriya.Say("Hello World!", voice="Daniel")
        >>> supriya.play(say)  # doctest: +SKIP

    """

    ### CLASS VARIABLES ###

    _voices = (
        "Alex",
        "Alice",
        "Alva",
        "Amelie",
        "Anna",
        "Carmit",
        "Damayanti",
        "Daniel",
        "Diego",
        "Ellen",
        "Fiona",
        "Fred",
        "Ioana",
        "Joana",
        "Jorge",
        "Juan",
        "Kanya",
        "Karen",
        "Kyoko",
        "Laura",
        "Lekha",
        "Luca",
        "Luciana",
        "Maged",
        "Mariska",
        "Mei-Jia",
        "Melina",
        "Milena",
        "Moira",
        "Monica",
        "Nora",
        "Paulina",
        "Samantha",
        "Sara",
        "Satu",
        "Sin-ji",
        "Tessa",
        "Thomas",
        "Ting-Ting",
        "Veena",
        "Victoria",
        "Xander",
        "Yelda",
        "Yuna",
        "Yuri",
        "Zosia",
        "Zuzana",
    )

    ### INITIALIZER ###

    def __init__(self, text, voice=None):
        self._text = str(text)
        if voice is not None:
            voice = str(voice)
            assert voice in self._voices
        self._voice = voice

    ### SPECIAL METHODS ###

    def __render__(
        self,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        print_transcript: bool = False,
        **kwargs,
    ) -> pathlib.Path:
        file_path = self._build_output_file_path(
            output_file_path=output_file_path,
            render_directory_path=render_directory_path,
        )
        assert file_path.parent.exists()
        if file_path.is_absolute():
            cwd = pathlib.Path.cwd()
            try:
                relative_file_path = file_path.relative_to(cwd)
            except ValueError:
                relative_file_path = file_path
        if print_transcript:
            print("Rendering {}".format(relative_file_path))
        if uqbar.io.find_executable("say"):
            command_parts = ["say"]
            command_parts.extend(["-o", str(relative_file_path)])
            if self.voice:
                command_parts.extend(["-v", self.voice])
        else:
            command_parts = ["espeak", "-w", str(relative_file_path)]
        command_parts.append(shlex.quote(self.text))
        command = " ".join(command_parts)
        if print_transcript:
            print("    Command: {}".format(command))
        if file_path.exists():
            if print_transcript:
                print(
                    "    Skipping {}. File already exists.".format(relative_file_path)
                )
            return file_path
        exit_code = subprocess.call(command, shell=True)
        if print_transcript:
            print(
                "    Rendered {} with exit code {}.".format(
                    relative_file_path, exit_code
                )
            )
        if exit_code:
            raise RuntimeError
        return file_path

    ### PRIVATE METHODS ###

    def _build_file_path(self):
        md5 = hashlib.md5()
        md5.update(self.text.encode())
        if self.voice is not None:
            md5.update(self.voice.encode())
        md5 = md5.hexdigest()
        file_path = "{}-{}.aiff".format(
            uqbar.strings.to_dash_case(type(self).__name__), md5
        )
        return pathlib.Path(file_path)

    def _build_output_file_path(
        self, output_file_path=None, render_directory_path=None
    ) -> pathlib.Path:
        if output_file_path:
            output_file_path = pathlib.Path(output_file_path).expanduser().absolute()
        elif render_directory_path:
            render_directory_path = (
                pathlib.Path(render_directory_path).expanduser().absolute()
            )
            output_file_path = render_directory_path / self._build_file_path()
        else:
            output_file_path = self._build_file_path()
            render_directory_path = (
                pathlib.Path(supriya.output_path).expanduser().absolute()
            )
            output_file_path = render_directory_path / self._build_file_path()
        return output_file_path

    ### PUBLIC PROPERTIES ###

    @property
    def text(self):
        return self._text

    @property
    def voice(self):
        return self._voice


class SoundFile(SupriyaObject):

    ### INITIALIZER ###

    def __init__(self, file_path):
        self._file_path = pathlib.Path(file_path)
        if not self._file_path.exists():
            raise ValueError(self._file_path)
        headers = sndhdr.what(self._file_path)
        self._frame_count = headers.nframes
        self._sample_rate = headers.framerate
        self._channel_count = headers.nchannels
        self._sample_width = headers.sampwidth
        self._file_type = headers.filetype

    ### PUBLIC METHODS ###

    def at_frame(self, frame):
        if self.file_type not in ("aiff", "wav"):
            raise ValueError(self.file_type)
        if not (0 <= frame <= self.frame_count):
            raise ValueError(frame)
        if self.sample_width not in (16, 24, 32):
            raise ValueError(f"Cannot decode sample width {self.sample_width}")
        with open(self.file_path, "rb") as file_pointer:
            reader_proc = {"aiff": aifc.open, "wav": wave.open}[self.file_type]
            with reader_proc(file_pointer, "rb") as reader:
                reader.setpos(frame)
                raw_data = reader.readframes(1)
        maximum = 2 ** (self.sample_width - 1)
        integers = []
        stride = self.sample_width // 8
        endianness = "big" if self.file_type == "aiff" else "little"
        for i in range(self.channel_count):
            chunk = raw_data[i * stride : i * stride + stride]
            integers.append(int.from_bytes(chunk, endianness, signed=True))
        return [float(x) / maximum for x in integers]

    def at_percent(self, percent):
        return self.at_frame(int(self.frame_count * percent))

    def at_second(self, second):
        return self.at_frame(int(second * self.sample_rate))

    ### PUBLIC PROPERTIES ###

    @property
    def channel_count(self):
        return self._channel_count

    @property
    def seconds(self):
        return float(self._frame_count) / float(self._sample_rate)

    @property
    def file_path(self):
        return self._file_path

    @property
    def file_type(self):
        return self._file_type

    @property
    def frame_count(self):
        return self._frame_count

    @property
    def sample_rate(self):
        return self._sample_rate

    @property
    def sample_width(self):
        return self._sample_width



import abc
import collections
from collections.abc import Iterable, Sequence

from supriya import CalculationRate, utils
from supriya.synthdefs import MultiOutUGen, PseudoUGen, UGen


class In(MultiOutUGen):
    """
    A bus input unit generator.

    ::

        >>> supriya.ugens.In.ar(bus=0, channel_count=4)
        UGenArray({4})

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _is_input = True
    _ordered_input_names = collections.OrderedDict([("bus", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class InFeedback(MultiOutUGen):
    """
    A bus input unit generator.

    Reads signal from a bus with a current or one cycle old timestamp.

    ::

        >>> in_feedback = supriya.ugens.InFeedback.ar(
        ...     bus=0,
        ...     channel_count=2,
        ... )
        >>> in_feedback
        UGenArray({2})

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _is_input = True
    _ordered_input_names = collections.OrderedDict([("bus", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LocalIn(MultiOutUGen):
    """
    A SynthDef-local bus input.

    ::

        >>> supriya.ugens.LocalIn.ar(channel_count=2)
        UGenArray({2})

    """

    ### CLASS VARIABLES ###

    _default_channel_count = 1
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict([("default", 0)])
    _unexpanded_input_names = ("default",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, channel_count=1, default=0):
        if not isinstance(default, Sequence):
            default = (default,)
        default = (float(_) for _ in default)
        default = utils.repeat_sequence_to_length(default, channel_count)
        default = list(default)[:channel_count]
        MultiOutUGen.__init__(
            self,
            calculation_rate=calculation_rate,
            channel_count=channel_count,
            default=default,
        )


class LocalOut(UGen):
    """
    A SynthDef-local bus output.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.LocalOut.ar(
        ...     source=source,
        ... )
        LocalOut.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class OffsetOut(UGen):
    """
    A bus output unit generator with sample-accurate timing.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.OffsetOut.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        OffsetOut.ar()

    """

    _default_channel_count = 0
    _is_output = True
    _ordered_input_names = collections.OrderedDict([("bus", 0), ("source", None)])
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Out(UGen):
    """
    A bus output unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.Out.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        Out.ar()

    """

    _default_channel_count = 0
    _is_output = True
    _ordered_input_names = collections.OrderedDict([("bus", 0.0), ("source", None)])
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class ReplaceOut(UGen):
    """
    An overwriting bus output unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.ReplaceOut.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        ReplaceOut.ar()

    """

    _default_channel_count = 0
    _is_output = True
    _ordered_input_names = collections.OrderedDict([("bus", 0), ("source", None)])
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class SoundIn(PseudoUGen):

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self):
        raise NotImplementedError

    ### PUBLIC METHODS ###

    @staticmethod
    def ar(bus=0):
        import supriya.ugens

        channel_offset = supriya.ugens.NumOutputBuses.ir()
        if isinstance(bus, Iterable):
            assert all(isinstance(x, int) for x in bus)
            bus = tuple(sorted(bus))
        else:
            assert isinstance(bus, int)
            bus = (bus,)
        if bus == tuple(range(min(bus), max(bus) + 1)):
            channel_count = len(bus)
            bus = min(bus)
        else:
            channel_count = 1
        bus = bus + channel_offset
        return supriya.ugens.In.ar(bus=bus, channel_count=channel_count)


class XOut(UGen):
    """
    A cross-fading bus output unit generator.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> xout = supriya.ugens.XOut.ar(
        ...     bus=0,
        ...     crossfade=0.5,
        ...     source=source,
        ... )
        >>> xout
        XOut.ar()

    """

    _default_channel_count = 0
    _is_output = True
    _ordered_input_names = collections.OrderedDict(
        [("bus", 0), ("crossfade", 0), ("source", None)]
    )
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



"""
Tools for modeling overlapping time structures with timespans.
"""
from .Interval import Interval  # noqa
from .IntervalTree import IntervalTree  # noqa
from .IntervalTreeDriver import IntervalTreeDriver  # noqa
from .Moment import Moment  # noqa

try:
    from .IntervalTreeDriverEx import IntervalTreeDriverEx  # noqa
except ModuleNotFoundError:
    pass



import collections

from supriya import CalculationRate, utils
from supriya.synthdefs import PseudoUGen, UGen


class Mix(PseudoUGen):
    """
    A down-to-mono signal mixer.

    ..  container:: example

        ::

            >>> with supriya.synthdefs.SynthDefBuilder() as builder:
            ...     oscillators = [supriya.ugens.DC.ar(1) for _ in range(5)]
            ...     mix = supriya.ugens.Mix.new(oscillators)
            ...
            >>> synthdef = builder.build(name="mix1", optimize=False)
            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> print(synthdef)
            synthdef:
                name: mix1
                ugens:
                -   DC.ar/0:
                        source: 1.0
                -   DC.ar/1:
                        source: 1.0
                -   DC.ar/2:
                        source: 1.0
                -   DC.ar/3:
                        source: 1.0
                -   Sum4.ar:
                        input_four: DC.ar/3[0]
                        input_one: DC.ar/0[0]
                        input_three: DC.ar/2[0]
                        input_two: DC.ar/1[0]
                -   DC.ar/4:
                        source: 1.0
                -   BinaryOpUGen(ADDITION).ar:
                        left: Sum4.ar[0]
                        right: DC.ar/4[0]

    ..  container:: example

        ::

            >>> with supriya.synthdefs.SynthDefBuilder() as builder:
            ...     oscillators = [supriya.ugens.DC.ar(1) for _ in range(15)]
            ...     mix = supriya.ugens.Mix.new(oscillators)
            ...
            >>> synthdef = builder.build("mix2")
            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> print(synthdef)
            synthdef:
                name: mix2
                ugens:
                -   DC.ar/0:
                        source: 1.0
                -   DC.ar/1:
                        source: 1.0
                -   DC.ar/2:
                        source: 1.0
                -   DC.ar/3:
                        source: 1.0
                -   Sum4.ar/0:
                        input_four: DC.ar/3[0]
                        input_one: DC.ar/0[0]
                        input_three: DC.ar/2[0]
                        input_two: DC.ar/1[0]
                -   DC.ar/4:
                        source: 1.0
                -   DC.ar/5:
                        source: 1.0
                -   DC.ar/6:
                        source: 1.0
                -   DC.ar/7:
                        source: 1.0
                -   Sum4.ar/1:
                        input_four: DC.ar/7[0]
                        input_one: DC.ar/4[0]
                        input_three: DC.ar/6[0]
                        input_two: DC.ar/5[0]
                -   DC.ar/8:
                        source: 1.0
                -   DC.ar/9:
                        source: 1.0
                -   DC.ar/10:
                        source: 1.0
                -   DC.ar/11:
                        source: 1.0
                -   Sum4.ar/2:
                        input_four: DC.ar/11[0]
                        input_one: DC.ar/8[0]
                        input_three: DC.ar/10[0]
                        input_two: DC.ar/9[0]
                -   DC.ar/12:
                        source: 1.0
                -   DC.ar/13:
                        source: 1.0
                -   DC.ar/14:
                        source: 1.0
                -   Sum3.ar:
                        input_one: DC.ar/12[0]
                        input_three: DC.ar/14[0]
                        input_two: DC.ar/13[0]
                -   Sum4.ar/3:
                        input_four: Sum3.ar[0]
                        input_one: Sum4.ar/0[0]
                        input_three: Sum4.ar/2[0]
                        input_two: Sum4.ar/1[0]

    """

    ### PRIVATE METHODS ###

    @classmethod
    def _flatten_sources(cls, sources):
        import supriya.synthdefs

        flattened_sources = []
        for source in sources:
            if isinstance(source, supriya.synthdefs.UGenArray):
                flattened_sources.extend(source)
            else:
                flattened_sources.append(source)
        return supriya.synthdefs.UGenArray(flattened_sources)

    ### PUBLIC METHODS ###

    @classmethod
    def new(cls, sources):
        import supriya.synthdefs
        import supriya.ugens

        sources = cls._flatten_sources(sources)
        summed_sources = []
        for part in utils.group_iterable_by_count(sources, 4):
            if len(part) == 4:
                summed_sources.append(supriya.ugens.Sum4(*part))
            elif len(part) == 3:
                summed_sources.append(supriya.ugens.Sum3(*part))
            elif len(part) == 2:
                summed_sources.append(part[0] + part[1])
            else:
                summed_sources.append(part[0])
        if len(summed_sources) == 1:
            return summed_sources[0]
        return Mix.new(summed_sources)

    @classmethod
    def multichannel(cls, sources, channel_count):
        """
        Segment by channel count and mix down in parallel.

        ..  container:: example

            Combine panner outputs, first with first, second with second, etc.

            ::

                >>> source = supriya.ugens.SinOsc.ar(frequency=[440, 660, 880])
                >>> panner = supriya.ugens.PanAz.ar(
                ...     channel_count=4,
                ...     source=source,
                ...     position=supriya.ugens.LFNoise2.kr(),
                ... )
                >>> mix = supriya.ugens.Mix.multichannel(panner, channel_count=4)
                >>> out = supriya.ugens.Out.ar(bus=0, source=mix)
                >>> supriya.graph(out)  # doctest: +SKIP

            ::

                >>> print(out)
                synthdef:
                    name: ...
                    ugens:
                    -   SinOsc.ar/0:
                            frequency: 440.0
                            phase: 0.0
                    -   LFNoise2.kr:
                            frequency: 500.0
                    -   PanAz.ar/0:
                            amplitude: 1.0
                            orientation: 0.5
                            position: LFNoise2.kr[0]
                            source: SinOsc.ar/0[0]
                            width: 2.0
                    -   SinOsc.ar/1:
                            frequency: 660.0
                            phase: 0.0
                    -   PanAz.ar/1:
                            amplitude: 1.0
                            orientation: 0.5
                            position: LFNoise2.kr[0]
                            source: SinOsc.ar/1[0]
                            width: 2.0
                    -   SinOsc.ar/2:
                            frequency: 880.0
                            phase: 0.0
                    -   PanAz.ar/2:
                            amplitude: 1.0
                            orientation: 0.5
                            position: LFNoise2.kr[0]
                            source: SinOsc.ar/2[0]
                            width: 2.0
                    -   Sum3.ar/0:
                            input_one: PanAz.ar/0[0]
                            input_three: PanAz.ar/2[0]
                            input_two: PanAz.ar/1[0]
                    -   Sum3.ar/1:
                            input_one: PanAz.ar/0[1]
                            input_three: PanAz.ar/2[1]
                            input_two: PanAz.ar/1[1]
                    -   Sum3.ar/2:
                            input_one: PanAz.ar/0[2]
                            input_three: PanAz.ar/2[2]
                            input_two: PanAz.ar/1[2]
                    -   Sum3.ar/3:
                            input_one: PanAz.ar/0[3]
                            input_three: PanAz.ar/2[3]
                            input_two: PanAz.ar/1[3]
                    -   Out.ar:
                            bus: 0.0
                            source[0]: Sum3.ar/0[0]
                            source[1]: Sum3.ar/1[0]
                            source[2]: Sum3.ar/2[0]
                            source[3]: Sum3.ar/3[0]

            Compare with a non-multichannel mixdown:

                >>> mix = supriya.ugens.Mix.new(panner)
                >>> out = supriya.ugens.Out.ar(bus=0, source=mix)
                >>> supriya.graph(out)  # doctest: +SKIP

            ::

                >>> print(out)
                synthdef:
                    name: ...
                    ugens:
                    -   SinOsc.ar/0:
                            frequency: 440.0
                            phase: 0.0
                    -   LFNoise2.kr:
                            frequency: 500.0
                    -   PanAz.ar/0:
                            amplitude: 1.0
                            orientation: 0.5
                            position: LFNoise2.kr[0]
                            source: SinOsc.ar/0[0]
                            width: 2.0
                    -   Sum4.ar/0:
                            input_four: PanAz.ar/0[3]
                            input_one: PanAz.ar/0[0]
                            input_three: PanAz.ar/0[2]
                            input_two: PanAz.ar/0[1]
                    -   SinOsc.ar/1:
                            frequency: 660.0
                            phase: 0.0
                    -   PanAz.ar/1:
                            amplitude: 1.0
                            orientation: 0.5
                            position: LFNoise2.kr[0]
                            source: SinOsc.ar/1[0]
                            width: 2.0
                    -   Sum4.ar/1:
                            input_four: PanAz.ar/1[3]
                            input_one: PanAz.ar/1[0]
                            input_three: PanAz.ar/1[2]
                            input_two: PanAz.ar/1[1]
                    -   SinOsc.ar/2:
                            frequency: 880.0
                            phase: 0.0
                    -   PanAz.ar/2:
                            amplitude: 1.0
                            orientation: 0.5
                            position: LFNoise2.kr[0]
                            source: SinOsc.ar/2[0]
                            width: 2.0
                    -   Sum4.ar/2:
                            input_four: PanAz.ar/2[3]
                            input_one: PanAz.ar/2[0]
                            input_three: PanAz.ar/2[2]
                            input_two: PanAz.ar/2[1]
                    -   Sum3.ar:
                            input_one: Sum4.ar/0[0]
                            input_three: Sum4.ar/2[0]
                            input_two: Sum4.ar/1[0]
                    -   Out.ar:
                            bus: 0.0
                            source[0]: Sum3.ar[0]

        """
        import supriya.synthdefs

        sources = cls._flatten_sources(sources)
        mixes, parts = [], []
        for i in range(0, len(sources), channel_count):
            parts.append(sources[i : i + channel_count])
        for columns in zip(*parts):
            mixes.append(cls.new(columns))
        return supriya.synthdefs.UGenArray(mixes)


class MulAdd(UGen):
    """
    An Optimized multiplication / addition ugen.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> mul_add = supriya.ugens.MulAdd.new(
        ...     addend=0.5,
        ...     multiplier=-1.5,
        ...     source=source,
        ... )
        >>> mul_add
        MulAdd.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("multiplier", 1.0), ("addend", 0.0)]
    )

    ### INITIALIZER ###

    def __init__(self, addend=0.0, multiplier=1.0, calculation_rate=None, source=None):
        UGen.__init__(
            self,
            addend=addend,
            multiplier=multiplier,
            calculation_rate=calculation_rate,
            source=source,
        )

    ### PRIVATE METHODS ###

    @staticmethod
    def _inputs_are_valid(source, multiplier, addend):
        if CalculationRate.from_expr(source) == CalculationRate.AUDIO:
            return True
        if CalculationRate.from_expr(source) == CalculationRate.CONTROL:
            if CalculationRate.from_expr(multiplier) in (
                CalculationRate.CONTROL,
                CalculationRate.SCALAR,
            ):
                if CalculationRate.from_expr(addend) in (
                    CalculationRate.CONTROL,
                    CalculationRate.SCALAR,
                ):
                    return True
        return False

    @classmethod
    def _new_single(
        cls, addend=None, multiplier=None, calculation_rate=None, source=None
    ):
        if multiplier == 0.0:
            return addend
        minus = multiplier == -1
        no_multiplier = multiplier == 1
        no_addend = addend == 0
        if no_multiplier and no_addend:
            return source
        if minus and no_addend:
            return -source
        if no_addend:
            return source * multiplier
        if minus:
            return addend - source
        if no_multiplier:
            return source + addend
        if cls._inputs_are_valid(source, multiplier, addend):
            return cls(
                addend=addend,
                multiplier=multiplier,
                calculation_rate=calculation_rate,
                source=source,
            )
        if cls._inputs_are_valid(multiplier, source, addend):
            return cls(
                addend=addend,
                multiplier=source,
                calculation_rate=calculation_rate,
                source=multiplier,
            )
        return (source * multiplier) + addend

    ### PUBLIC METHODS ###

    @classmethod
    def new(cls, source=None, multiplier=1.0, addend=0.0):
        """
        Constructs a multiplication / addition ugen.

        ::

            >>> addend = 0.5
            >>> multiplier = 1.5
            >>> source = supriya.ugens.SinOsc.ar(frequency=[440, 442])
            >>> mul_add = supriya.ugens.MulAdd.new(
            ...     addend=addend,
            ...     multiplier=multiplier,
            ...     source=source,
            ... )
            >>> mul_add
            UGenArray({2})

        Returns ugen graph.
        """
        import supriya.synthdefs

        # TODO: handle case of array as source
        calculation_rate = supriya.CalculationRate.from_expr(
            (source, multiplier, addend)
        )
        ugen = cls._new_expanded(
            addend=addend,
            multiplier=multiplier,
            calculation_rate=calculation_rate,
            source=source,
        )
        return ugen


class Sum3(UGen):
    """
    A three-input summing unit generator.

    ::

        >>> input_one = supriya.ugens.SinOsc.ar()
        >>> input_two = supriya.ugens.SinOsc.ar(phase=0.1)
        >>> input_three = supriya.ugens.SinOsc.ar(phase=0.2)
        >>> supriya.ugens.Sum3.new(
        ...     input_one=input_one,
        ...     input_two=input_two,
        ...     input_three=input_three,
        ... )
        Sum3.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [("input_one", None), ("input_two", None), ("input_three", None)]
    )

    _valid_calculation_rates = ()

    ### INITIALIZER ###

    def __init__(self, input_one=None, input_two=None, input_three=None):
        inputs = [input_one, input_two, input_three]
        calculation_rate = CalculationRate.from_expr(inputs)
        inputs.sort(key=lambda x: CalculationRate.from_expr(x), reverse=True)
        inputs = tuple(inputs)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            input_one=input_one,
            input_two=input_two,
            input_three=input_three,
        )

    ### PRIVATE METHODS ###

    @classmethod
    def _new_single(cls, input_one=None, input_two=None, input_three=None, **kwargs):
        if input_three == 0:
            ugen = input_one + input_two
        elif input_two == 0:
            ugen = input_one + input_three
        elif input_one == 0:
            ugen = input_two + input_three
        else:
            ugen = cls(
                input_one=input_one, input_two=input_two, input_three=input_three
            )
        return ugen


class Sum4(UGen):
    """
    A four-input summing unit generator.

    ::

        >>> input_one = supriya.ugens.SinOsc.ar()
        >>> input_two = supriya.ugens.SinOsc.ar(phase=0.1)
        >>> input_three = supriya.ugens.SinOsc.ar(phase=0.2)
        >>> input_four = supriya.ugens.SinOsc.ar(phase=0.3)
        >>> supriya.ugens.Sum4.new(
        ...     input_one=input_one,
        ...     input_two=input_two,
        ...     input_three=input_three,
        ...     input_four=input_four,
        ... )
        Sum4.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [
            ("input_one", None),
            ("input_two", None),
            ("input_three", None),
            ("input_four", None),
        ]
    )

    _valid_calculation_rates = ()

    ### INITIALIZER ###

    def __init__(
        self, input_one=None, input_two=None, input_three=None, input_four=None
    ):
        inputs = [input_one, input_two, input_three, input_four]
        calculation_rate = CalculationRate.from_expr(inputs)
        inputs.sort(key=lambda x: CalculationRate.from_expr(x), reverse=True)
        inputs = tuple(inputs)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            input_one=input_one,
            input_two=input_two,
            input_three=input_three,
            input_four=input_four,
        )

    ### PRIVATE METHODS ###

    @classmethod
    def _new_single(
        cls, input_one=None, input_two=None, input_three=None, input_four=None, **kwargs
    ):
        import supriya.ugens

        if input_one == 0:
            ugen = supriya.ugens.Sum3.new(
                input_one=input_two, input_two=input_three, input_three=input_four
            )
        elif input_two == 0:
            ugen = supriya.ugens.Sum3.new(
                input_one=input_one, input_two=input_three, input_three=input_four
            )
        elif input_three == 0:
            ugen = supriya.ugens.Sum3.new(
                input_one=input_one, input_two=input_two, input_three=input_four
            )
        elif input_four == 0:
            ugen = supriya.ugens.Sum3.new(
                input_one=input_one, input_two=input_two, input_three=input_three
            )
        else:
            ugen = cls(
                input_one=input_one,
                input_two=input_two,
                input_three=input_three,
                input_four=input_four,
            )
        return ugen



import bisect

import supriya.commands
from supriya.nonrealtime.bases import SessionObject


class Buffer(SessionObject):
    """
    A non-realtime buffer.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_buffer_group",
        "_channel_count",
        "_duration",
        "_events",
        "_frame_count",
        "_session",
        "_session_id",
        "_start_offset",
        "_file_path",
        "_starting_frame",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        session,
        session_id,
        buffer_group=None,
        channel_count=None,
        frame_count=1,
        duration=None,
        start_offset=None,
        file_path=None,
        starting_frame=None,
    ):
        SessionObject.__init__(self, session)
        self._events = {}
        self._session_id = int(session_id)
        if buffer_group is not None:
            assert isinstance(buffer_group, BufferGroup)
        self._buffer_group = buffer_group
        start_offset = start_offset or 0
        self._start_offset = float(start_offset)
        if duration is None:
            duration = float("inf")
        self._duration = duration
        if file_path is not None:
            starting_frame = int(starting_frame or 0)
        else:
            channel_count = int(channel_count or 1)
            frame_count = int(frame_count or 1)
        self._frame_count = frame_count
        self._channel_count = channel_count
        self._starting_frame = starting_frame
        self._file_path = file_path

    ### PRIVATE METHODS ###

    @SessionObject.require_offset
    def _set_event(self, item, value, offset=None):
        if offset < 0 or self.duration < offset:
            return
        events = self._events.setdefault(item, [])
        new_event = (offset, value)
        if not events:
            events.append(new_event)
            return
        index = bisect.bisect_left(events, new_event)
        if len(events) <= index:
            events.append(new_event)
        old_offset, old_value = events[index]
        if old_offset == offset:
            events[index] = (offset, value)
        else:
            events.insert(index, new_event)

    ### PUBLIC METHODS ###

    def close(self, offset=None):
        event_type = supriya.commands.BufferCloseRequest
        event_kwargs = dict(buffer_id=self)
        self._set_event(event_type, event_kwargs, offset=offset)

    def copy(
        self,
        target_buffer_id,
        frame_count=None,
        source_starting_frame=None,
        target_starting_frame=None,
        offset=None,
    ):
        event_type = supriya.commands.BufferCopyRequest
        event_kwargs = dict(
            frame_count=frame_count,
            source_buffer_id=self,
            source_starting_frame=source_starting_frame,
            target_buffer_id=target_buffer_id,
            target_starting_frame=target_starting_frame,
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def fill(self, index_count_value_triples=None, offset=None):
        event_type = supriya.commands.BufferFillRequest
        event_kwargs = dict(
            buffer_id=self, index_count_value_triples=index_count_value_triples
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def fill_via_chebyshev(
        self,
        amplitudes,
        as_wavetable=True,
        should_normalize=True,
        should_clear_first=True,
        offset=None,
    ):
        event_type = supriya.commands.BufferGenerateRequest
        event_kwargs = dict(
            amplitudes=tuple(float(_) for _ in amplitudes),
            as_wavetable=bool(as_wavetable),
            buffer_id=self,
            command_name="cheby",
            should_clear_first=bool(should_clear_first),
            should_normalize=bool(should_normalize),
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def fill_via_sine_1(
        self,
        amplitudes,
        as_wavetable=True,
        should_clear_first=True,
        should_normalize=True,
        offset=None,
    ):
        event_type = supriya.commands.BufferGenerateRequest
        event_kwargs = dict(
            amplitudes=tuple(float(_) for _ in amplitudes),
            as_wavetable=bool(as_wavetable),
            buffer_id=self,
            command_name="sine1",
            should_clear_first=bool(should_clear_first),
            should_normalize=bool(should_normalize),
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def fill_via_sine_2(
        self,
        amplitudes,
        frequencies,
        as_wavetable=True,
        should_clear_first=True,
        should_normalize=True,
        offset=None,
    ):
        event_type = supriya.commands.BufferGenerateRequest
        event_kwargs = dict(
            amplitudes=tuple(float(_) for _ in amplitudes),
            as_wavetable=bool(as_wavetable),
            buffer_id=self,
            command_name="sine2",
            frequencies=tuple(float(_) for _ in frequencies),
            should_clear_first=bool(should_clear_first),
            should_normalize=bool(should_normalize),
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def fill_via_sine_3(
        self,
        amplitudes,
        frequencies,
        phases,
        as_wavetable=True,
        should_clear_first=True,
        should_normalize=True,
        offset=None,
    ):
        event_type = supriya.commands.BufferGenerateRequest
        event_kwargs = dict(
            amplitudes=tuple(float(_) for _ in amplitudes),
            as_wavetable=bool(as_wavetable),
            buffer_id=self,
            command_name="sine3",
            frequencies=tuple(float(_) for _ in frequencies),
            phases=tuple(float(_) for _ in phases),
            should_clear_first=bool(should_clear_first),
            should_normalize=bool(should_normalize),
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def normalize(
        self, as_wavetable=None, buffer_id=None, new_maximum=1.0, offset=None
    ):
        event_type = supriya.commands.BufferNormalizeRequest
        event_kwargs = dict(
            buffer_id=self, new_maximum=new_maximum, as_wavetable=as_wavetable
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def read(
        self,
        file_path,
        channel_indices=None,
        frame_count=None,
        leave_open=None,
        starting_frame_in_buffer=None,
        starting_frame_in_file=None,
        offset=None,
    ):
        event_type = supriya.commands.BufferReadRequest
        # need to optionally coerce to BufferReadChannelRequest on compile
        event_kwargs = dict(
            buffer_id=self,
            file_path=file_path,
            channel_indices=channel_indices,
            frame_count=frame_count,
            leave_open=leave_open,
            starting_frame_in_buffer=starting_frame_in_buffer,
            starting_frame_in_file=starting_frame_in_file,
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def set(self, index_value_pairs=None, offset=None):
        event_type = supriya.commands.BufferSetRequest
        event_kwargs = dict(buffer_id=self, index_value_pairs=index_value_pairs)
        self._set_event(event_type, event_kwargs, offset=offset)

    def set_contiguous(self, index_values_pairs=None, offset=None):
        event_type = supriya.commands.BufferSetContiguousRequest
        event_kwargs = dict(buffer_id=self, index_values_pairs=index_values_pairs)
        self._set_event(event_type, event_kwargs, offset=offset)

    def write(
        self,
        file_path,
        frame_count=None,
        header_format="aiff",
        leave_open=False,
        sample_format="int24",
        starting_frame=None,
        offset=None,
    ):
        event_type = supriya.commands.BufferWriteRequest
        event_kwargs = dict(
            buffer_id=self,
            file_path=file_path,
            frame_count=frame_count,
            header_format=header_format,
            leave_open=leave_open,
            sample_format=sample_format,
            starting_frame=starting_frame,
        )
        self._set_event(event_type, event_kwargs, offset=offset)

    def zero(self, offset=None):
        event_type = supriya.commands.BufferZeroRequest
        event_kwargs = dict(buffer_id=self)
        self._set_event(event_type, event_kwargs, offset=offset)

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_group(self):
        return self._buffer_group

    @property
    def channel_count(self):
        return self._channel_count

    @property
    def duration(self):
        return self._duration

    @property
    def frame_count(self):
        return self._frame_count

    @property
    def session_id(self):
        return self._session_id

    @property
    def start_offset(self):
        return self._start_offset

    @property
    def starting_frame(self):
        return self._starting_frame

    @property
    def file_path(self):
        return self._file_path

    @property
    def stop_offset(self):
        if self.duration is None:
            return None
        return self.start_offset + self.duration


class BufferGroup(SessionObject):
    """
    A non-realtime buffer group.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_buffers", "_session")

    ### INITIALIZER ###

    def __init__(
        self,
        session,
        buffer_count=1,
        duration=None,
        channel_count=1,
        frame_count=1,
        start_offset=None,
    ):
        import supriya.nonrealtime

        SessionObject.__init__(self, session)
        buffer_count = int(buffer_count)
        assert 0 < buffer_count
        buffers = []
        start_id = len(self.session.buffers)
        for session_id in range(start_id, buffer_count + start_id):
            buffer_ = supriya.nonrealtime.Buffer(
                session,
                session_id=session_id,
                buffer_group=self,
                duration=duration,
                channel_count=channel_count,
                frame_count=frame_count,
                start_offset=start_offset,
            )
            buffers.append(buffer_)
        self._buffers = tuple(buffers)

    ### SPECIAL METHODS ###

    def __contains__(self, item):
        return self.buffers.__contains__(item)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._buffers[item]
        elif isinstance(item, slice):
            return tuple(self._buffers[item])

    def __iter__(self):
        for buffer in self._buffers:
            yield buffer

    def __len__(self):
        return len(self._buffers)

    ### PUBLIC METHODS ###

    def index(self, buffer_):
        return self._buffers.index(buffer_)

    ### PUBLIC PROPERTIES ###

    @property
    def buffers(self):
        return self._buffers

    @property
    def buffer_count(self):
        return len(self)

    @property
    def channel_count(self):
        return self._buffers[0].channel_count

    @property
    def duration(self):
        return self._buffers[0].duration

    @property
    def frame_count(self):
        return self._buffers[0].frame_count

    @property
    def start_offset(self):
        return self._buffers[0].start_offset

    @property
    def stop_offset(self):
        return self._buffers[0].stop_offset



import collections

from supriya import CalculationRate
from supriya.synthdefs import PseudoUGen, PureUGen


class Filter(PureUGen):
    """
    Abstract base class for filter ugens.
    """

    def _validate_inputs(self):
        self._check_rate_same_as_first_input_rate()


class APF(Filter):
    """
    An all-pass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> apf = supriya.ugens.APF.ar(
        ...     frequency=440,
        ...     radius=0.8,
        ...     source=source,
        ... )
        >>> apf
        APF.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440.0), ("radius", 0.8)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BPF(Filter):
    """
    A 2nd order Butterworth bandpass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> b_p_f = supriya.ugens.BPF.ar(source=source)
        >>> b_p_f
        BPF.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440.0), ("reciprocal_of_q", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BPZ2(Filter):
    """
    A two zero fixed midpass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> bpz_2 = supriya.ugens.BPZ2.ar(
        ...     source=source,
        ... )
        >>> bpz_2
        BPZ2.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BRF(Filter):
    """
    A 2nd order Butterworth band-reject filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> b_r_f = supriya.ugens.BRF.ar(source=source)
        >>> b_r_f
        BRF.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440.0), ("reciprocal_of_q", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BRZ2(Filter):
    """
    A two zero fixed midcut filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> brz_2 = supriya.ugens.BRZ2.ar(
        ...     source=source,
        ... )
        >>> brz_2
        BRZ2.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Changed(PseudoUGen):
    """
    Triggers when a value changes.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> changed = supriya.ugens.Changed.ar(
        ...     source=source,
        ...     threshold=0,
        ... )
        >>> supriya.graph(changed)  # doctest: +SKIP

    ::

        >>> print(changed)
        synthdef:
            name: 39e1f9d61589c4acaaf297cc961d65e4
            ugens:
            -   In.ar:
                    bus: 0.0
            -   HPZ1.ar:
                    source: In.ar[0]
            -   UnaryOpUGen(ABSOLUTE_VALUE).ar:
                    source: HPZ1.ar[0]
            -   BinaryOpUGen(GREATER_THAN).ar:
                    left: UnaryOpUGen(ABSOLUTE_VALUE).ar[0]
                    right: 0.0

    """

    ### PUBLIC METHODS ###

    @classmethod
    def ar(cls, source=None, threshold=0):
        """
        Constructs an audio-rate Changed.

        ::

            >>> source = supriya.ugens.In.ar(bus=0)
            >>> changed = supriya.ugens.Changed.ar(
            ...     source=source,
            ...     threshold=0,
            ... )
            >>> supriya.graph(changed)  # doctest: +SKIP

        ::

            >>> print(changed)
            synthdef:
                name: 39e1f9d61589c4acaaf297cc961d65e4
                ugens:
                -   In.ar:
                        bus: 0.0
                -   HPZ1.ar:
                        source: In.ar[0]
                -   UnaryOpUGen(ABSOLUTE_VALUE).ar:
                        source: HPZ1.ar[0]
                -   BinaryOpUGen(GREATER_THAN).ar:
                        left: UnaryOpUGen(ABSOLUTE_VALUE).ar[0]
                        right: 0.0

        Returns ugen graph.
        """
        import supriya.ugens

        ugen = abs(supriya.ugens.HPZ1.ar(source=source)) > threshold
        return ugen

    @classmethod
    def kr(cls, source=None, threshold=0):
        """
        Constructs a control-rate Changed.

        ::

            >>> source = supriya.ugens.In.ar(bus=0)
            >>> changed = supriya.ugens.Changed.kr(
            ...     source=source,
            ...     threshold=0,
            ... )
            >>> supriya.graph(changed)  # doctest: +SKIP

        ::

            >>> print(changed)
            synthdef:
                name: e2436271176995c6a0a5cac6d1553f8b
                ugens:
                -   In.ar:
                        bus: 0.0
                -   HPZ1.kr:
                        source: In.ar[0]
                -   UnaryOpUGen(ABSOLUTE_VALUE).kr:
                        source: HPZ1.kr[0]
                -   BinaryOpUGen(GREATER_THAN).kr:
                        left: UnaryOpUGen(ABSOLUTE_VALUE).kr[0]
                        right: 0.0

        Returns ugen graph.
        """
        import supriya.ugens

        ugen = abs(supriya.ugens.HPZ1.kr(source=source)) > threshold
        return ugen


class Decay(Filter):
    """
    A leaky signal integrator.

    ::

        >>> source = supriya.ugens.Impulse.ar()
        >>> decay = supriya.ugens.Decay.ar(
        ...     source=source,
        ... )
        >>> decay
        Decay.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("decay_time", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Decay2(Filter):
    """
    A leaky signal integrator.

    ::

        >>> source = supriya.ugens.Impulse.ar()
        >>> decay_2 = supriya.ugens.Decay2.ar(
        ...     source=source,
        ... )
        >>> decay_2
        Decay2.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("attack_time", 0.01), ("decay_time", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class DetectSilence(Filter):
    """
    Evaluates `done_action` when input falls below `threshold`.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> source *= supriya.ugens.Line.kr(start=1, stop=0)
        >>> detect_silence = supriya.ugens.DetectSilence.kr(
        ...     done_action=supriya.DoneAction.FREE_SYNTH,
        ...     source=source,
        ...     threshold=0.0001,
        ...     time=1.0,
        ... )
        >>> detect_silence
        DetectSilence.kr()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("threshold", 0.0001), ("time", 0.1), ("done_action", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### PRIVATE METHODS ###

    def _optimize_graph(self, sort_bundles):
        # TODO: Replace with `_is_pure = False` class variable
        pass


class FOS(Filter):
    """
    A first order filter section.

    ::

        out(i) = (a0 * in(i)) + (a1 * in(i-1)) + (a2 * in(i-2)) + (b1 * out(i-1)) + (b2 * out(i-2))

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> fos = supriya.ugens.FOS.ar(
        ...     a_0=0,
        ...     a_1=0,
        ...     b_1=0,
        ...     source=source,
        ... )
        >>> fos
        FOS.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("a_0", 0.0), ("a_1", 0.0), ("b_1", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Formlet(Filter):
    """
    A FOF-like filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> formlet = supriya.ugens.Formlet.ar(
        ...     attack_time=1,
        ...     decay_time=1,
        ...     frequency=440,
        ...     source=source,
        ... )
        >>> formlet
        Formlet.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("frequency", 440.0),
            ("attack_time", 1.0),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class HPF(Filter):
    """
    A Highpass filter unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.HPF.ar(source=source)
        HPF.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class HPZ1(Filter):
    """
    A two point difference filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> hpz_1 = supriya.ugens.HPZ1.ar(
        ...     source=source,
        ... )
        >>> hpz_1
        HPZ1.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class HPZ2(Filter):
    """
    A two zero fixed midcut filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> hpz_2 = supriya.ugens.HPZ2.ar(
        ...     source=source,
        ... )
        >>> hpz_2
        HPZ2.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Integrator(Filter):
    """
    A leaky integrator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> integrator = supriya.ugens.Integrator.ar(
        ...     coefficient=1,
        ...     source=source,
        ... )
        >>> integrator
        Integrator.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("coefficient", 1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Lag(Filter):
    """
    A lag generator.

    ::

        >>> source = supriya.ugens.In.kr(bus=0)
        >>> supriya.ugens.Lag.kr(
        ...     lag_time=0.5,
        ...     source=source,
        ... )
        Lag.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("lag_time", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LagUD(Filter):
    """
    An up/down lag generator.

    ::

        >>> source = supriya.ugens.In.kr(bus=0)
        >>> supriya.ugens.LagUD.kr(
        ...     lag_time_down=1.25,
        ...     lag_time_up=0.5,
        ...     source=source,
        ... )
        LagUD.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("lag_time_up", 0.1), ("lag_time_down", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Lag2(Filter):
    """
    An exponential lag generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lag_2 = supriya.ugens.Lag2.ar(
        ...     lag_time=0.1,
        ...     source=source,
        ... )
        >>> lag_2
        Lag2.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("lag_time", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Lag2UD(Filter):
    """
    An up/down exponential lag generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lag_2_ud = supriya.ugens.Lag2UD.ar(
        ...     lag_time_d=0.1,
        ...     lag_time_u=0.1,
        ...     source=source,
        ... )
        >>> lag_2_ud
        Lag2UD.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("lag_time_u", 0.1), ("lag_time_d", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Lag3(Filter):
    """
    An exponential lag generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lag_3 = supriya.ugens.Lag3.ar(
        ...     lag_time=0.1,
        ...     source=source,
        ... )
        >>> lag_3
        Lag3.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("lag_time", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Lag3UD(Filter):
    """
    An up/down exponential lag generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lag_3_ud = supriya.ugens.Lag3UD.ar(
        ...     lag_time_d=0.1,
        ...     lag_time_u=0.1,
        ...     source=source,
        ... )
        >>> lag_3_ud
        Lag3UD.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("lag_time_u", 0.1), ("lag_time_d", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LeakDC(Filter):
    """
    A DC blocker.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> leak_d_c = supriya.ugens.LeakDC.ar(
        ...     source=source,
        ...     coefficient=0.995,
        ... )
        >>> leak_d_c
        LeakDC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("coefficient", 0.995)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LPF(Filter):
    """
    A lowpass filter unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.LPF.ar(source=source)
        LPF.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LPZ1(Filter):
    """
    A two point average filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lpz_1 = supriya.ugens.LPZ1.ar(
        ...     source=source,
        ... )
        >>> lpz_1
        LPZ1.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LPZ2(Filter):
    """
    A two zero fixed lowpass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lpz_2 = supriya.ugens.LPZ2.ar(
        ...     source=source,
        ... )
        >>> lpz_2
        LPZ2.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Median(Filter):
    """
    A median filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> median = supriya.ugens.Median.ar(
        ...     length=3,
        ...     source=source,
        ... )
        >>> median
        Median.ar()

    """

    _ordered_input_names = collections.OrderedDict([("length", 3), ("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class MidEQ(Filter):
    """
    A parametric filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> mid_eq = supriya.ugens.MidEQ.ar(
        ...     db=0,
        ...     frequency=440,
        ...     reciprocal_of_q=1,
        ...     source=source,
        ... )
        >>> mid_eq
        MidEQ.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440), ("reciprocal_of_q", 1), ("db", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class MoogFF(Filter):
    """
    A Moog VCF implementation.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> moog_ff = supriya.ugens.MoogFF.ar(
        ...     frequency=100,
        ...     gain=2,
        ...     reset=0,
        ...     source=source,
        ... )
        >>> moog_ff
        MoogFF.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 100), ("gain", 2), ("reset", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class OnePole(Filter):
    """
    A one pole filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> one_pole = supriya.ugens.OnePole.ar(
        ...     coefficient=0.5,
        ...     source=source,
        ... )
        >>> one_pole
        OnePole.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("coefficient", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class OneZero(Filter):
    """
    A one zero filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> one_zero = supriya.ugens.OneZero.ar(
        ...     coefficient=0.5,
        ...     source=source,
        ... )
        >>> one_zero
        OneZero.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("coefficient", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class RHPF(Filter):
    """
    A resonant highpass filter unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.RLPF.ar(source=source)
        RLPF.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440), ("reciprocal_of_q", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class RLPF(Filter):
    """
    A resonant lowpass filter unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.RLPF.ar(source=source)
        RLPF.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440), ("reciprocal_of_q", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Ramp(Filter):
    """
    Breaks a continuous signal into line segments.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> ramp = supriya.ugens.Ramp.ar(
        ...     lag_time=0.1,
        ...     source=source,
        ... )
        >>> ramp
        Ramp.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("lag_time", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Ringz(Filter):
    """
    A ringing filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> ringz = supriya.ugens.Ringz.ar(
        ...     decay_time=1,
        ...     frequency=440,
        ...     source=source,
        ... )
        >>> ringz
        Ringz.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440), ("decay_time", 1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class SOS(Filter):
    """
    A second-order filter section.

    ::

        out(i) = (a0 * in(i)) + (a1 * in(i-1)) + (b1 * out(i-1))

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> sos = supriya.ugens.SOS.ar(
        ...     a_0=0,
        ...     a_1=0,
        ...     a_2=0,
        ...     b_1=0,
        ...     b_2=0,
        ...     source=source,
        ... )
        >>> sos
        SOS.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("a_0", 0), ("a_1", 0), ("a_2", 0), ("b_1", 0), ("b_2", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Slew(Filter):
    """
    A slew rate limiter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> slew = supriya.ugens.Slew.ar(
        ...     source=source,
        ...     up=1,
        ...     down=1,
        ... )
        >>> slew
        Slew.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("up", 1), ("down", 1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Slope(Filter):
    """
    Calculates slope of signal.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> slope = supriya.ugens.Slope.ar(
        ...     source=source,
        ... )
        >>> slope
        Slope.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class TwoPole(Filter):
    """
    A two pole filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> two_pole = supriya.ugens.TwoPole.ar(
        ...     frequency=440,
        ...     radius=0.8,
        ...     source=source,
        ... )
        >>> two_pole
        TwoPole.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440), ("radius", 0.8)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class TwoZero(Filter):
    """
    A two zero filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> two_zero = supriya.ugens.TwoZero.ar(
        ...     frequency=440,
        ...     radius=0.8,
        ...     source=source,
        ... )
        >>> two_zero
        TwoZero.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 440), ("radius", 0.8)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen


class FreeVerb(UGen):
    """
    A FreeVerb reverb unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.FreeVerb.ar(
        ...     source=source,
        ... )
        FreeVerb.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("mix", 0.33), ("room_size", 0.5), ("damping", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)



from collections.abc import Sequence

from supriya import ParameterRate
from supriya.system import SupriyaValueObject


class QueryTreeControl(SupriyaValueObject):

    ### CLASS VARIABLES ###

    __slots__ = ("_control_value", "_control_name_or_index")

    ### INITIALIZER ###

    def __init__(self, control_name_or_index=None, control_value=None):
        self._control_value = control_value
        self._control_name_or_index = control_name_or_index

    ### SPECIAL METHODS ###

    def __str__(self):
        key = self._control_name_or_index
        value = self._control_value
        try:
            value = round(value, 6)
        except Exception:
            pass
        string = "{}: {!s}".format(key, value)
        return string

    ### PUBLIC METHODS ###

    @classmethod
    def from_control(cls, control):
        import supriya.realtime

        control_name = control.name
        if isinstance(control.value, supriya.realtime.Bus):
            control_value = str(control.value)
        else:
            control_value = float(control.value)
        return cls(control_value=control_value, control_name_or_index=control_name)

    ### PUBLIC PROPERTIES ###

    @property
    def control_name_or_index(self):
        return self._control_name_or_index

    @property
    def control_value(self):
        return self._control_value


class QueryTreeSynth(SupriyaValueObject, Sequence):

    ### CLASS VARIABLES ###

    __slots__ = ("_controls", "_extra", "_name", "_node_id", "_synthdef_name")

    ### INITIALIZER ###

    def __init__(
        self, node_id=None, synthdef_name=None, controls=None, name=None, **extra
    ):
        self._controls = controls
        self._extra = tuple(sorted(extra.items()))
        self._node_id = node_id
        self._synthdef_name = synthdef_name
        self._name = name

    ### SPECIAL METHODS ###

    def __format__(self, format_spec):
        result = self._get_str_format_pieces(unindexed=format_spec == "unindexed")
        result = "\n".join(result)
        return result

    def __getitem__(self, item):
        return self._controls[item]

    def __len__(self):
        return len(self._controls)

    def __str__(self):
        result = self._get_str_format_pieces()
        result = "\n".join(result)
        return result

    ### PRIVATE METHODS ###

    @classmethod
    def _from_nrt_synth(
        cls,
        state,
        node,
        include_controls=False,
        include_timespans=False,
        id_mapping=None,
    ):
        from supriya.nonrealtime import Bus, BusGroup, Synth
        from supriya.synthdefs import SynthDef

        assert isinstance(node, Synth)
        node_id = node.session_id
        synthdef_name = node.synthdef
        if isinstance(synthdef_name, SynthDef):
            synthdef_name = synthdef_name.actual_name
        controls = []
        if include_controls:
            settings = node._collect_settings(
                state.offset, persistent=True, id_mapping=id_mapping
            )
            synthdef, synth_kwargs = node.synthdef, node.synth_kwargs
            for name, parameter in sorted(synthdef.parameters.items()):
                value = parameter.value
                if name in synth_kwargs:
                    value = synth_kwargs[name]
                if name in settings:
                    value = settings[name]
                if (
                    parameter.parameter_rate == ParameterRate.SCALAR
                    or parameter.name in ("in_", "out")
                ):
                    if value in id_mapping:
                        value = id_mapping[value]
                    value = float(value)
                elif isinstance(value, (Bus, BusGroup)) and value in id_mapping:
                    value = value.get_map_symbol(id_mapping[value])
                else:
                    value = float(value)
                control = QueryTreeControl(
                    control_name_or_index=name, control_value=value
                )
                controls.append(control)
        extra = {}
        if include_timespans:
            extra.update(timespan=[node.start_offset, node.stop_offset])
        query_tree_synth = QueryTreeSynth(
            node_id=node_id, synthdef_name=synthdef_name, controls=controls, **extra
        )
        return query_tree_synth

    def _get_str_format_pieces(self, unindexed=False):
        result = []
        node_id = self.node_id
        if unindexed:
            node_id = "..."
        string = f"{node_id} {self.synthdef_name}"
        if self.name:
            string += f" ({self.name})"
        if self.extra:
            string += (
                " ("
                + ", ".join("{}: {}".format(key, value) for key, value in self.extra)
                + ")"
            )
        result.append(string)
        if self.controls:
            control_string = ", ".join(str(control) for control in self.controls)
            control_string = "    " + control_string
            result.append(control_string)
        return result

    ### PUBLIC METHODS ###

    def annotate(self, annotation_map):
        return type(self)(
            node_id=self.node_id,
            controls=self.controls,
            synthdef_name=self.synthdef_name,
            name=annotation_map.get(self.node_id),
            **dict(self.extra),
        )

    @classmethod
    def from_response(cls, response):
        return cls(
            node_id=response.node_id,
            controls=[
                QueryTreeControl(control_name_or_index=name, control_value=value)
                for name, value in (response.synthdef_controls.items() or {})
            ],
            synthdef_name=response.synthdef_name,
        )

    @classmethod
    def from_synth(cls, synth, include_controls=False):
        import supriya.commands
        import supriya.realtime
        import supriya.synthdefs

        assert isinstance(synth, supriya.realtime.Synth)
        node_id = synth.node_id
        synthdef_name = synth.synthdef
        if isinstance(synthdef_name, supriya.synthdefs.SynthDef):
            synthdef_name = synthdef_name.actual_name
        controls = []
        if include_controls:
            for control_name in synth:
                control = QueryTreeControl.from_control(synth.controls[control_name])
                controls.append(control)
        query_tree_synth = QueryTreeSynth(
            node_id=node_id, synthdef_name=synthdef_name, controls=tuple(controls)
        )
        return query_tree_synth

    def to_dict(self):
        """
        Convert QueryTreeSynth to JSON-serializable dictionary.

        ::

            >>> query_tree_synth = supriya.querytree.QueryTreeSynth(
            ...     node_id=1001,
            ...     synthdef_name="c1aa521afab5b0c0ce3d744690951649",
            ...     controls=(
            ...         supriya.querytree.QueryTreeControl(
            ...             control_name_or_index="level",
            ...             control_value=1.0,
            ...         ),
            ...         supriya.querytree.QueryTreeControl(
            ...             control_name_or_index="out",
            ...             control_value=0.0,
            ...         ),
            ...     ),
            ... )

        ::

            >>> import json
            >>> result = query_tree_synth.to_dict()
            >>> result = json.dumps(
            ...     result,
            ...     indent=4,
            ...     separators=(",", ": "),
            ...     sort_keys=True,
            ... )
            >>> print(result)
            {
                "controls": {
                    "level": 1.0,
                    "out": 0.0
                },
                "node_id": 1001,
                "synthdef": "c1aa521afab5b0c0ce3d744690951649"
            }

        """
        result = {
            "node_id": self.node_id,
            "synthdef": self.synthdef_name,
            "controls": {},
        }
        for control in self.controls:
            name = control.control_name_or_index
            result["controls"][name] = control.control_value
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def controls(self):
        return self._controls

    @property
    def extra(self):
        return self._extra

    @property
    def name(self):
        return self._name

    @property
    def node_id(self):
        return self._node_id

    @property
    def synthdef_name(self):
        return self._synthdef_name


class QueryTreeGroup(SupriyaValueObject, Sequence):

    ### CLASS VARIABLES ###

    __slots__ = ("_children", "_extra", "_name", "_node_id")

    ### INITIALIZER ###

    def __init__(self, node_id=None, children=None, name=None, **extra):
        self._children = tuple(children or ())
        self._extra = tuple(sorted(extra.items()))
        self._node_id = node_id
        self._name = name

    ### SPECIAL METHODS ###

    def __format__(self, format_spec):
        result = self._get_str_format_pieces(unindexed=format_spec == "unindexed")
        result = "\n".join(result)
        result = "NODE TREE {}".format(result)
        return result

    def __getitem__(self, item):
        return self._children[item]

    def __len__(self):
        return len(self._children)

    def __str__(self):
        result = self._get_str_format_pieces()
        result = "\n".join(result)
        result = "NODE TREE {}".format(result)
        return result

    ### PRIVATE METHODS ###

    @classmethod
    def _from_nrt_group(
        cls,
        state,
        node,
        include_controls=False,
        include_timespans=False,
        id_mapping=None,
    ):
        import supriya.commands
        import supriya.nonrealtime

        assert isinstance(node, supriya.nonrealtime.Group)
        node_id = node.session_id
        children = []
        for child in state.nodes_to_children.get(node) or ():
            if isinstance(child, supriya.nonrealtime.Group):
                child = QueryTreeGroup._from_nrt_group(
                    state,
                    child,
                    include_controls=include_controls,
                    include_timespans=include_timespans,
                    id_mapping=id_mapping,
                )
            elif isinstance(child, supriya.nonrealtime.Synth):
                child = QueryTreeSynth._from_nrt_synth(
                    state,
                    child,
                    include_controls=include_controls,
                    include_timespans=include_timespans,
                    id_mapping=id_mapping,
                )
            else:
                raise ValueError(child)
            children.append(child)
        children = tuple(children)
        extra = {}
        if include_timespans:
            extra.update(timespan=[node.start_offset, node.stop_offset])
        query_tree_group = QueryTreeGroup(node_id=node_id, children=children, **extra)
        return query_tree_group

    def _get_str_format_pieces(self, unindexed=False):
        result = []
        node_id = self.node_id
        if unindexed:
            node_id = "..."
        string = f"{node_id} group"
        if self.name:
            string += f" ({self.name})"
        if self.extra:
            string += (
                " ("
                + ", ".join("{}: {}".format(key, value) for key, value in self.extra)
                + ")"
            )
        result.append(string)
        for child in self.children:
            for line in child._get_str_format_pieces(unindexed=unindexed):
                result.append("    {}".format(line))
        return result

    ### PUBLIC METHODS ###

    def annotate(self, annotation_map):
        return type(self)(
            node_id=self.node_id,
            children=[child.annotate(annotation_map) for child in self.children],
            name=annotation_map.get(self.node_id),
            **dict(self.extra),
        )

    @classmethod
    def from_group(cls, group, include_controls=False):
        import supriya.commands
        import supriya.realtime

        assert isinstance(group, supriya.realtime.Group)
        node_id = group.node_id
        children = []
        for child in group.children:
            if isinstance(child, supriya.realtime.Group):
                child = QueryTreeGroup.from_group(
                    child, include_controls=include_controls
                )
            elif isinstance(child, supriya.realtime.Synth):
                child = QueryTreeSynth.from_synth(
                    child, include_controls=include_controls
                )
            else:
                raise ValueError(child)
            children.append(child)
        children = tuple(children)
        query_tree_group = QueryTreeGroup(node_id=node_id, children=children)
        return query_tree_group

    @classmethod
    def from_response(cls, response):
        return cls(node_id=response.node_id)

    @classmethod
    def from_state(cls, state, include_controls=False, include_timespans=False):
        id_mapping = state.session._build_id_mapping()
        root_node = state.session.root_node
        query_tree_group = cls._from_nrt_group(
            state,
            root_node,
            include_controls=include_controls,
            include_timespans=include_timespans,
            id_mapping=id_mapping,
        )
        return query_tree_group

    def to_dict(self):
        """
        Convert QueryTreeGroup to JSON-serializable dictionary.

        ::

            >>> query_tree_group = supriya.querytree.QueryTreeGroup(
            ...     node_id=1002,
            ...     children=(
            ...         supriya.querytree.QueryTreeSynth(
            ...             node_id=1105,
            ...             synthdef_name="dca557070c6b57164557041ac746fb3f",
            ...             controls=(
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="damping",
            ...                     control_value=0.06623425334692,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="duration",
            ...                     control_value=3.652155876159668,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="level",
            ...                     control_value=0.894767701625824,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="out",
            ...                     control_value=16.0,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="room_size",
            ...                     control_value=0.918643176555634,
            ...                 ),
            ...             ),
            ...         ),
            ...         supriya.querytree.QueryTreeSynth(
            ...             node_id=1098,
            ...             synthdef_name="cc754c63533fdcf412a44ef6adb1a8f0",
            ...             controls=(
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="duration",
            ...                     control_value=5.701356887817383,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="level",
            ...                     control_value=0.959683060646057,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="out",
            ...                     control_value=16.0,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="pitch_dispersion",
            ...                     control_value=0.040342573076487,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="pitch_shift",
            ...                     control_value=10.517594337463379,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="time_dispersion",
            ...                     control_value=0.666014134883881,
            ...                 ),
            ...                 supriya.querytree.QueryTreeControl(
            ...                     control_name_or_index="window_size",
            ...                     control_value=1.014111995697021,
            ...                 ),
            ...             ),
            ...         ),
            ...     ),
            ... )

        ::

            >>> import json
            >>> result = query_tree_group.to_dict()
            >>> result = json.dumps(
            ...     result,
            ...     indent=4,
            ...     separators=(",", ": "),
            ...     sort_keys=True,
            ... )
            >>> print(result)
            {
                "children": [
                    {
                        "controls": {
                            "damping": 0.06623425334692,
                            "duration": 3.652155876159668,
                            "level": 0.894767701625824,
                            "out": 16.0,
                            "room_size": 0.918643176555634
                        },
                        "node_id": 1105,
                        "synthdef": "dca557070c6b57164557041ac746fb3f"
                    },
                    {
                        "controls": {
                            "duration": 5.701356887817383,
                            "level": 0.959683060646057,
                            "out": 16.0,
                            "pitch_dispersion": 0.040342573076487,
                            "pitch_shift": 10.517594337463379,
                            "time_dispersion": 0.666014134883881,
                            "window_size": 1.014111995697021
                        },
                        "node_id": 1098,
                        "synthdef": "cc754c63533fdcf412a44ef6adb1a8f0"
                    }
                ],
                "node_id": 1002
            }

        """
        result = {
            "node_id": self.node_id,
            "children": [x.to_dict() for x in self.children],
        }
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def children(self):
        return self._children

    @property
    def extra(self):
        return self._extra

    @property
    def name(self):
        return self._name

    @property
    def node_id(self):
        return self._node_id



import abc
import collections
import contextlib
import dataclasses
import pathlib
import re
import tempfile
from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from uqbar.objects import new

import supriya.nonrealtime  # noqa
import supriya.realtime  # noqa
from supriya import commands, nonrealtime, realtime
from supriya.assets.synthdefs.default import default
from supriya.enums import AddAction, CalculationRate, ParameterRate
from supriya.nonrealtime import Session
from supriya.realtime import AsyncServer, BaseServer, Server
from supriya.synthdefs import SynthDef

# with provider.at(): proxy = provider.add_buffer(file_path=file_path)
# with provider.at(): proxy.free()


@dataclasses.dataclass(frozen=True)
class Proxy:
    provider: "Provider"


@dataclasses.dataclass(frozen=True)
class BufferProxy:
    provider: "Provider"
    identifier: Union["supriya.nonrealtime.Buffer", int]
    channel_count: Optional[int] = None
    frame_count: Optional[int] = None
    file_path: Optional[str] = None
    starting_frame: Optional[int] = None

    def __float__(self):
        return float(int(self))

    def __int__(self):
        if self.provider.server:
            return self.identifier
        elif self.provider.session:
            return self.provider.identifier.session_id

    def close(self):
        pass

    def free(self):
        self.provider.free_buffer(self)

    def normalize(self, new_maximum=1.0):
        pass

    def read(self, file_path, leave_open=False):
        pass

    def write(
        self,
        file_path,
        frame_count=None,
        header_format="aiff",
        leave_open=False,
        sample_format="int24",
        starting_frame=None,
    ):
        pass

    def as_allocate_request(self):
        kwargs = dict(buffer_id=int(self), frame_count=self.frame_count)
        if self.file_path is None:
            return commands.BufferAllocateRequest(
                **kwargs, channel_count=self.channel_count
            )
        kwargs.update(file_path=self.file_path, starting_frame=self.starting_frame)
        if self.channel_count is None:
            return commands.BufferAllocateReadRequest(**kwargs)
        return commands.BufferAllocateReadChannelRequest(
            **kwargs, channel_indices=list(range(self.channel_count))
        )

    def as_free_request(self):
        return commands.BufferFreeRequest(buffer_id=int(self))


@dataclasses.dataclass(frozen=True)
class OscCallbackProxy(Proxy):
    provider: "Provider"
    identifier: Any

    def unregister(self):
        self.provider.unregister_osc_callback(self)


@dataclasses.dataclass(frozen=True)
class BusProxy(Proxy):
    calculation_rate: CalculationRate
    provider: "Provider"
    identifier: Union["supriya.nonrealtime.Bus", int]

    def __float__(self):
        return float(int(self))

    def __int__(self):
        if self.provider.server:
            return self.identifier
        elif self.provider.session:
            return self.provider.identifier.session_id

    def set_(self, value):
        self.provider.set_bus(self, value)

    def free(self):
        self.provider.free_bus(self)

    @property
    def map_symbol(self):
        if self.calculation_rate == CalculationRate.AUDIO:
            return f"a{int(self)}"
        return f"c{int(self)}"


@dataclasses.dataclass(frozen=True)
class BusGroupProxy(Proxy):
    calculation_rate: CalculationRate
    channel_count: int
    identifier: Union["supriya.nonrealtime.BusGroup", int]
    provider: "Provider"
    buses: Sequence["BusProxy"] = dataclasses.field(init=False)

    def __post_init__(self):
        if isinstance(self.identifier, int):
            bus_identifiers = range(
                self.identifier, self.identifier + self.channel_count
            )
        else:
            bus_identifiers = self.identifier[:]
        object.__setattr__(
            self,
            "buses",
            tuple(
                BusProxy(
                    calculation_rate=self.calculation_rate,
                    provider=self.provider,
                    identifier=bus_identifier,
                )
                for bus_identifier in bus_identifiers
            ),
        )

    def __float__(self):
        return float(int(self))

    def __getitem__(self, item):
        return self.buses[item]

    def __int__(self):
        if self.provider.server:
            return self.identifier
        elif self.provider.session:
            return self.provider.identifier.session_id

    def __len__(self):
        return self.channel_count

    def free(self):
        self.provider.free_bus_group(self)


@dataclasses.dataclass(frozen=True)
class NodeProxy(Proxy):
    identifier: Union["supriya.nonrealtime.Node", int]
    provider: "Provider"

    def __float__(self):
        return float(int(self))

    def __int__(self):
        if self.provider.server:
            return self.identifier
        elif self.provider.session:
            return self.provider.identifier.session_id

    def __setitem__(self, key, value):
        self.provider.set_node(self, **{key: value})

    def add_group(
        self, *, add_action: int = AddAction.ADD_TO_HEAD, name: Optional[str] = None
    ) -> "GroupProxy":
        return self.provider.add_group(add_action=add_action, target_node=self)

    def add_synth(
        self,
        *,
        synthdef: SynthDef = None,
        add_action: int = AddAction.ADD_TO_HEAD,
        name: Optional[str] = None,
        **settings,
    ) -> "SynthProxy":
        return self.provider.add_synth(
            add_action=add_action, synthdef=synthdef, target_node=self, **settings
        )

    def as_move_request(
        self, add_action: AddAction, target_node: "NodeProxy"
    ) -> commands.MoveRequest:
        request_classes: Dict[int, Type[commands.MoveRequest]] = {
            AddAction.ADD_TO_HEAD: commands.GroupHeadRequest,
            AddAction.ADD_TO_TAIL: commands.GroupTailRequest,
            AddAction.ADD_BEFORE: commands.NodeBeforeRequest,
            AddAction.ADD_AFTER: commands.NodeAfterRequest,
        }
        request_class: Type[commands.MoveRequest] = request_classes[add_action]
        return request_class(
            node_id_pairs=[request_class.NodeIdPair(int(self), int(target_node))]
        )

    def as_set_request(self, **settings):
        coerced_settings = {}
        for key, value in settings.items():
            if isinstance(value, (BusProxy, BusGroupProxy)):
                if value.calculation_rate == CalculationRate.AUDIO:
                    value = f"a{value.identifier}"
                else:
                    value = f"c{value.identifier}"
            coerced_settings[key] = value
        return commands.NodeSetRequest(node_id=int(self), **coerced_settings)

    def dispose(self):
        self.provider.dispose(self)

    def free(self):
        self.provider.free_node(self)

    def move(self, add_action: AddAction, target_node: "NodeProxy"):
        self.provider.move_node(self, add_action, target_node)


@dataclasses.dataclass(frozen=True)
class GroupProxy(NodeProxy):
    identifier: Union["supriya.nonrealtime.Node", int]
    provider: "Provider"

    def as_add_request(self, add_action, target_node):
        return commands.GroupNewRequest(
            items=[
                commands.GroupNewRequest.Item(
                    node_id=int(self.identifier),
                    add_action=add_action,
                    target_node_id=int(target_node),
                )
            ]
        )

    def as_free_request(self, force=False):
        return commands.NodeFreeRequest(node_ids=[int(self)])


@dataclasses.dataclass(frozen=True)
class SynthProxy(NodeProxy):
    identifier: Union["supriya.nonrealtime.Node", int]
    provider: "Provider"
    synthdef: SynthDef
    settings: Dict[str, Union[float, BusGroupProxy]]

    def as_add_request(self, add_action, target_node):
        # TODO: Handle map symbols
        #       If arg is a bus proxy, and synth param is scalar, cast to int
        #       Elif arg is a bus proxy, and synth param not scalar, map
        #       Else cast to float
        synthdef = self.synthdef or default

        synthdef_kwargs = {}
        for _, parameter in synthdef.indexed_parameters:
            if parameter.name not in self.settings:
                continue
            value = self.settings[parameter.name]
            if value == parameter.value:
                continue
            if parameter.parameter_rate == ParameterRate.SCALAR:
                synthdef_kwargs[parameter.name] = float(value)
            elif parameter.name in ("in_", "out"):
                synthdef_kwargs[parameter.name] = float(value)
            elif isinstance(value, (BusProxy, BusGroupProxy)):
                synthdef_kwargs[parameter.name] = value.map_symbol
            else:
                synthdef_kwargs[parameter.name] = float(value)

        return commands.SynthNewRequest(
            node_id=int(self.identifier),
            add_action=add_action,
            target_node_id=int(target_node),
            synthdef=synthdef,
            **synthdef_kwargs,
        )

    def as_free_request(self, force=False):
        if force or "gate" not in self.synthdef.parameters:
            return commands.NodeFreeRequest(node_ids=[int(self)])
        return commands.NodeSetRequest(node_id=int(self), gate=0)


@dataclasses.dataclass(frozen=True)
class ProviderMoment:
    provider: "Provider"
    seconds: float
    bus_settings: List[Tuple[BusProxy, float]] = dataclasses.field(default_factory=list)
    buffer_additions: List[BufferProxy] = dataclasses.field(default_factory=list)
    buffer_removals: List[BufferProxy] = dataclasses.field(default_factory=list)
    node_reorderings: List[Tuple[NodeProxy, AddAction, NodeProxy]] = dataclasses.field(
        default_factory=list
    )
    node_additions: List[Tuple[NodeProxy, AddAction, NodeProxy]] = dataclasses.field(
        default_factory=list
    )
    node_removals: List[NodeProxy] = dataclasses.field(default_factory=list)
    node_settings: List[
        Tuple[NodeProxy, Dict[str, Union[float, BusGroupProxy]]]
    ] = dataclasses.field(default_factory=list)
    wait: bool = dataclasses.field(default=False)
    exit_stack: contextlib.ExitStack = dataclasses.field(
        init=False, default_factory=contextlib.ExitStack, compare=False
    )

    async def __aenter__(self):
        if self.provider.server and not isinstance(self.provider.server, AsyncServer):
            raise RuntimeError(repr(self.provider.server))
        return self._enter()

    async def __aexit__(self, *args):
        results = self._exit()
        if not results:
            return
        timestamp, request_bundle, synthdefs = results
        server = self.provider.server
        # The underlying asyncio UDP transport will silently drop oversize packets
        if len(request_bundle.to_datagram()) <= 8192:
            if self.wait:
                # If waiting, the original ProviderMoment timestamp can be ignored
                await request_bundle.communicate_async(server=server, sync=True)
            else:
                server.send(request_bundle.to_osc())
        else:
            # If over the UDP packet limit, partition the message
            requests = request_bundle.contents
            # Always wait for SynthDefs to load.
            if synthdefs:
                synthdef_request = requests[0]
                requests = synthdef_request.callback.contents or []
                synthdef_request = new(synthdef_request, callback=None)
                await synthdef_request.communicate_async(sync=True, server=server)
            if self.wait:
                # If waiting, the original ProviderMoment timestamp can be ignored
                for bundle in commands.RequestBundle.partition(requests):
                    await bundle.communicate_async(server=server, sync=True)
            else:
                for bundle in commands.RequestBundle.partition(
                    requests, timestamp=timestamp
                ):
                    server.send(bundle.to_osc())

    def __enter__(self):
        if self.provider.session is not None:
            self.exit_stack.enter_context(self.provider.session.at(self.seconds or 0))
        if self.provider.server and not isinstance(self.provider.server, Server):
            raise RuntimeError(repr(self.provider.server))
        return self._enter()

    def __exit__(self, *args):
        results = self._exit()
        if not results:
            return
        timestamp, request_bundle, synthdefs = results
        try:
            self.provider.server.send(request_bundle.to_osc())
        except OSError:
            requests = request_bundle.contents
            if synthdefs:
                synthdef_request = requests[0]
                requests = synthdef_request.callback.contents or []
                synthdef_request = new(synthdef_request, callback=None)
                synthdef_request.communicate(sync=True, server=self.provider.server)
            for bundle in commands.RequestBundle.partition(
                requests, timestamp=timestamp
            ):
                self.provider.server.send(bundle.to_osc())

    def _enter(self):
        self.provider._moments.append(self)
        self.provider._counter[self.seconds] += 1
        return self

    def _exit(self):
        self.exit_stack.close()
        self.provider._moments.pop()
        self.provider._counter[self.seconds] -= 1
        if not self.provider.server:
            return
        elif self.provider._counter[self.seconds]:
            return
        requests = []
        synthdefs = set()
        new_nodes = set()
        for buffer_proxy in self.buffer_additions:
            requests.append(buffer_proxy.as_allocate_request())
        for node_proxy, add_action, target_node in self.node_additions:
            request = node_proxy.as_add_request(add_action, target_node)
            if isinstance(request, commands.SynthNewRequest):
                if request.synthdef not in self.provider.server:
                    synthdefs.add(request.synthdef)
            requests.append(request)
            new_nodes.add(node_proxy.identifier)
        for node_proxy, add_action, target_node in self.node_reorderings:
            requests.append(node_proxy.as_move_request(add_action, target_node))
        for node_proxy, settings in self.node_settings:
            requests.append(node_proxy.as_set_request(**settings))
        for node_proxy in self.node_removals:
            requests.append(
                node_proxy.as_free_request(force=node_proxy.identifier in new_nodes)
            )
        for buffer_proxy in self.buffer_removals:
            requests.append(buffer_proxy.as_free_request())
        if self.bus_settings:
            sorted_pairs = sorted(
                dict(
                    (int(bus_proxy.identifier), value)
                    for bus_proxy, value in self.bus_settings
                ).items()
            )
            request = commands.ControlBusSetRequest(index_value_pairs=sorted_pairs)
            requests.append(request)
        if not requests:
            return
        timestamp = self.seconds
        if timestamp is not None:
            timestamp += self.provider._latency
        if synthdefs:
            request_bundle = commands.RequestBundle(
                timestamp=timestamp,
                contents=[
                    commands.SynthDefReceiveRequest(
                        synthdefs=sorted(synthdefs, key=lambda x: x.actual_name),
                        callback=commands.RequestBundle(contents=requests),
                    )
                ],
            )
            # check bundle size, write synthdefs to disk and do /d_load
            if len(request_bundle.to_datagram(with_placeholders=True)) > 8192:
                directory_path = pathlib.Path(tempfile.mkdtemp())
                # directory_path = pathlib.Path("~/Desktop").expanduser()
                for synthdef in synthdefs:
                    name = synthdef.anonymous_name
                    if synthdef.name:
                        name += "-" + re.sub(r"[^\w]", "-", synthdef.name)
                    file_name = "{}.scsyndef".format(name)
                    synthdef_path = directory_path / file_name
                    synthdef_path.write_bytes(synthdef.compile())
                request_bundle = commands.RequestBundle(
                    timestamp=timestamp,
                    contents=[
                        supriya.commands.SynthDefLoadDirectoryRequest(
                            directory_path=directory_path,
                            callback=commands.RequestBundle(contents=requests),
                        )
                    ],
                )
        else:
            request_bundle = commands.RequestBundle(
                timestamp=timestamp, contents=requests
            )
        for synthdef in synthdefs:
            synthdef._register_with_local_server(server=self.provider.server)
        return timestamp, request_bundle, synthdefs


class Provider(metaclass=abc.ABCMeta):
    """
    Provides limited realtime/non-realtime compatibility layer.
    """

    ### INITIALIZER ###

    def __init__(self, latency=0.1):
        self._moments: List[ProviderMoment] = []
        self._counter = collections.Counter()
        self._server = None
        self._session = None
        self._latency = latency
        self._annotation_map: Dict[Union["supriya.nonrealtime.Node", int], str] = {}

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def add_buffer(
        self,
        *,
        channel_count: Optional[int] = None,
        file_path: Optional[str] = None,
        frame_count: Optional[int] = None,
        starting_frame: Optional[int] = None,
    ) -> BufferProxy:
        raise NotImplementedError

    @abc.abstractmethod
    def add_bus(self, calculation_rate=CalculationRate.CONTROL) -> BusProxy:
        raise NotImplementedError

    @abc.abstractmethod
    def add_bus_group(
        self, channel_count=1, calculation_rate=CalculationRate.CONTROL
    ) -> BusGroupProxy:
        raise NotImplementedError

    @abc.abstractmethod
    def add_group(
        self,
        *,
        target_node=None,
        add_action=AddAction.ADD_TO_HEAD,
        name: Optional[str] = None,
    ) -> GroupProxy:
        raise NotImplementedError

    @abc.abstractmethod
    def add_synth(
        self,
        *,
        synthdef: SynthDef = None,
        target_node=None,
        add_action=AddAction.ADD_TO_HEAD,
        name: Optional[str] = None,
        **settings,
    ) -> SynthProxy:
        raise NotImplementedError

    @abc.abstractmethod
    def boot(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def dispose(self, node_proxy: NodeProxy):
        raise NotImplementedError

    @abc.abstractmethod
    def free_buffer(self, buffer_proxy):
        raise NotImplementedError

    @abc.abstractmethod
    def free_bus(self, bus_proxy: BusProxy):
        raise NotImplementedError

    @abc.abstractmethod
    def free_bus_group(self, bus_group_proxy: BusGroupProxy):
        raise NotImplementedError

    @abc.abstractmethod
    def free_node(self, node_proxy: NodeProxy):
        raise NotImplementedError

    @abc.abstractmethod
    def move_node(
        self, node_proxy: NodeProxy, add_action: AddAction, target_node: NodeProxy
    ):
        raise NotImplementedError

    @abc.abstractmethod
    def set_bus(self, bus_proxy: BusProxy, value: float):
        raise NotImplementedError

    @abc.abstractmethod
    def set_node(self, node_proxy: NodeProxy, **settings):
        raise NotImplementedError

    def at(self, seconds=None, wait=False):
        if self._moments and self._moments[-1].seconds == seconds:
            provider_moment = self._moments[-1]
        else:
            provider_moment = ProviderMoment(provider=self, seconds=seconds, wait=wait)
        return provider_moment

    @classmethod
    def from_context(cls, context, latency=0.1) -> "Provider":
        if isinstance(context, Session):
            return NonrealtimeProvider(context, latency=latency)
        elif isinstance(context, BaseServer):
            return RealtimeProvider(context, latency=latency)
        raise ValueError("Unknown context")

    @classmethod
    def nonrealtime(cls) -> "NonrealtimeProvider":
        session = Session()
        return cast("NonrealtimeProvider", cls.from_context(session))

    @abc.abstractmethod
    def quit(self):
        raise NotImplementedError

    @classmethod
    def realtime(
        cls, scsynth_path=None, options=None, port=None, **kwargs
    ) -> "RealtimeProvider":
        server = Server()
        server.boot(port=port, scsynth_path=scsynth_path, options=options, **kwargs)
        return cast("RealtimeProvider", cls.from_context(server))

    @classmethod
    async def realtime_async(
        cls, scsynth_path=None, options=None, port=None, **kwargs
    ) -> "RealtimeProvider":
        server = AsyncServer()
        await server.boot(
            port=port, scsynth_path=scsynth_path, options=options, **kwargs
        )
        return cast("RealtimeProvider", cls.from_context(server))

    @abc.abstractmethod
    def register_osc_callback(
        self, pattern: Tuple[Union[str, float], ...], procedure: Callable
    ) -> OscCallbackProxy:
        raise NotImplementedError

    @abc.abstractmethod
    def unregister_osc_callback(self, proxy: OscCallbackProxy):
        raise NotImplementedError

    ### PUBLIC PROPERTIES ###

    @property
    def annotation_map(self) -> Mapping[Union["supriya.nonrealtime.Node", int], str]:
        return MappingProxyType(self._annotation_map)

    @property
    def latency(self):
        return self._latency

    @property
    def moment(self) -> Optional[ProviderMoment]:
        if self._moments:
            return self._moments[-1]
        return None

    @property
    def server(self) -> Server:
        return self._server

    @property
    def session(self) -> Session:
        return self._session


class NonrealtimeProvider(Provider):

    ### INITIALIZER ###

    def __init__(self, session, latency=0.1):
        if not isinstance(session, Session):
            raise ValueError(f"Expected session, got {session}")
        Provider.__init__(self, latency=latency)
        self._session = session

    ### SPECIAL METHODS ###

    def __str__(self):
        return f"<{type(self).__name__} {self._session!r}>"

    ### PRIVATE METHODS ###

    def _resolve_target_node(self, target_node) -> nonrealtime.Node:
        if target_node is None:
            target_node = self.session.root_node
        elif isinstance(target_node, NodeProxy):
            target_node = target_node.identifier
        return target_node

    ### PUBLIC METHODS ###

    def add_buffer(
        self,
        *,
        channel_count: Optional[int] = None,
        file_path: Optional[str] = None,
        frame_count: Optional[int] = None,
        starting_frame: Optional[int] = None,
    ) -> BufferProxy:
        if not self.moment:
            raise ValueError("No current moment")
        identifier = self.session.add_buffer(
            channel_count=channel_count,
            file_path=file_path,
            frame_count=frame_count,
            starting_frame=starting_frame,
        )
        return BufferProxy(
            channel_count=channel_count,
            file_path=file_path,
            frame_count=frame_count,
            identifier=identifier,
            provider=self,
            starting_frame=starting_frame,
        )

    def add_bus(self, calculation_rate=CalculationRate.CONTROL) -> BusProxy:
        if not self.moment:
            raise ValueError("No current moment")
        calculation_rate = CalculationRate.from_expr(calculation_rate)
        if calculation_rate not in (CalculationRate.AUDIO, CalculationRate.CONTROL):
            raise ValueError(f"Invalid calculation rate: {calculation_rate!r}")
        identifier = self.session.add_bus(calculation_rate=calculation_rate)
        return BusProxy(
            calculation_rate=calculation_rate, identifier=identifier, provider=self
        )

    def add_bus_group(
        self, channel_count=1, calculation_rate=CalculationRate.CONTROL
    ) -> BusGroupProxy:
        if not self.moment:
            raise ValueError("No current moment")
        calculation_rate = CalculationRate.from_expr(calculation_rate)
        if calculation_rate not in (CalculationRate.AUDIO, CalculationRate.CONTROL):
            raise ValueError(f"Invalid calculation rate: {calculation_rate!r}")
        if channel_count < 1:
            raise ValueError("Channel-count must be positive, non-zero integer")
        identifier = self.session.add_bus_group(
            bus_count=channel_count, calculation_rate=calculation_rate
        )
        return BusGroupProxy(
            calculation_rate=calculation_rate,
            channel_count=channel_count,
            identifier=identifier,
            provider=self,
        )

    def add_group(
        self,
        *,
        target_node=None,
        add_action=AddAction.ADD_TO_HEAD,
        name: Optional[str] = None,
    ) -> GroupProxy:
        if not self.moment:
            raise ValueError("No current moment")
        identifier = self._resolve_target_node(target_node).add_group(
            add_action=add_action
        )
        proxy = GroupProxy(identifier=identifier, provider=self)
        return proxy

    def add_synth(
        self,
        *,
        synthdef: SynthDef = None,
        target_node=None,
        add_action=AddAction.ADD_TO_HEAD,
        name: Optional[str] = None,
        **settings,
    ) -> SynthProxy:
        if not self.moment:
            raise ValueError("No current moment")
        sanitized_settings = {}
        for key, value in settings.items():
            if isinstance(value, (BusProxy, BusGroupProxy)):
                value = value.identifier
            sanitized_settings[key] = value
        identifier = self._resolve_target_node(target_node).add_synth(
            add_action=add_action, synthdef=synthdef, **sanitized_settings
        )
        proxy = SynthProxy(
            identifier=identifier,
            provider=self,
            synthdef=synthdef or default,
            settings=settings,
        )
        return proxy

    def free_buffer(self, buffer_: BufferProxy):
        if not self.moment:
            raise ValueError("No current moment")
        return  # This is currently a no-op

    def boot(self, **kwargs):
        pass  # no-op

    def dispose(self, node_proxy: NodeProxy):
        if not self.moment:
            raise ValueError("No current moment")
        return  # This is currently a no-op

    def free_bus(self, bus: BusProxy):
        if not self.moment:
            raise ValueError("No current moment")
        return  # This is currently a no-op

    def free_bus_group(self, bus_group: BusGroupProxy):
        if not self.moment:
            raise ValueError("No current moment")
        return  # This is currently a no-op

    def free_node(self, node_proxy: NodeProxy):
        if not self.moment:
            raise ValueError("No current moment")
        cast(nonrealtime.Node, node_proxy.identifier).free()

    def move_node(
        self,
        node_proxy: NodeProxy,
        add_action: AddAction,
        target_node: Union[NodeProxy, nonrealtime.Node],
    ):
        if not self.moment:
            raise ValueError("No current moment")
        self._resolve_target_node(target_node).move_node(
            node_proxy.identifier, add_action=add_action
        )

    def set_bus(self, bus_proxy: BusProxy, value: float):
        if not self.moment:
            raise ValueError("No current moment")
        elif bus_proxy.calculation_rate != CalculationRate.CONTROL:
            raise ValueError("Can only set control-rate buses")
        cast(nonrealtime.Bus, bus_proxy.identifier).set_(value)

    def set_node(self, node_proxy: NodeProxy, **settings):
        if not self.moment:
            raise ValueError("No current moment")
        for key, value in settings.items():
            if isinstance(value, (BusProxy, BusGroupProxy)):
                value = value.identifier
            cast(nonrealtime.Node, node_proxy.identifier)[key] = value

    def quit(self):
        pass  # no-op

    def register_osc_callback(
        self, pattern: Tuple[Union[str, float], ...], procedure: Callable
    ) -> OscCallbackProxy:
        return OscCallbackProxy(provider=self, identifier=None)

    def unregister_osc_callback(self, proxy: OscCallbackProxy):
        pass  # no-op


class RealtimeProvider(Provider):

    ### INITIALIZER ###

    def __init__(self, server, latency=0.1):
        if not isinstance(server, BaseServer):
            raise ValueError(f"Expected Server, got {server}")
        Provider.__init__(self, latency=latency)
        self._server = server

    ### SPECIAL METHODS ###

    def __str__(self):
        return f"<{type(self).__name__} {self._server!r}>"

    ### PRIVATE METHODS ###

    def _resolve_target_node(self, target_node):
        if target_node is None:
            # TODO: Will this work with AsyncServer?
            target_node = self.server.default_group
        return target_node

    ### PUBLIC METHODS ###

    def add_buffer(
        self,
        *,
        channel_count: Optional[int] = None,
        file_path: Optional[str] = None,
        frame_count: Optional[int] = None,
        starting_frame: Optional[int] = None,
    ) -> BufferProxy:
        if not self.moment:
            raise ValueError("No current moment")
        identifier = self.server.buffer_allocator.allocate(1)
        proxy = BufferProxy(
            channel_count=channel_count,
            file_path=file_path,
            frame_count=frame_count,
            identifier=identifier,
            provider=self,
            starting_frame=starting_frame,
        )
        self.moment.buffer_additions.append(proxy)
        return proxy

    def add_bus(self, calculation_rate=CalculationRate.CONTROL) -> BusProxy:
        if not self.moment:
            raise ValueError("No current moment")
        calculation_rate = CalculationRate.from_expr(calculation_rate)
        if calculation_rate not in (CalculationRate.AUDIO, CalculationRate.CONTROL):
            raise ValueError(f"Invalid calculation rate: {calculation_rate!r}")
        allocator = realtime.Bus._get_allocator(calculation_rate, server=self.server)
        identifier = allocator.allocate(1)
        return BusProxy(
            calculation_rate=calculation_rate, identifier=identifier, provider=self
        )

    def add_bus_group(
        self, channel_count=1, calculation_rate=CalculationRate.CONTROL
    ) -> BusGroupProxy:
        if not self.moment:
            raise ValueError("No current moment")
        calculation_rate = CalculationRate.from_expr(calculation_rate)
        if calculation_rate not in (CalculationRate.AUDIO, CalculationRate.CONTROL):
            raise ValueError(f"Invalid calculation rate: {calculation_rate!r}")
        if channel_count < 1:
            raise ValueError("Channel-count must be positive, non-zero integer")
        allocator = realtime.Bus._get_allocator(calculation_rate, server=self.server)
        identifier = allocator.allocate(channel_count)
        if identifier is None:
            raise RuntimeError
        return BusGroupProxy(
            calculation_rate=calculation_rate,
            channel_count=channel_count,
            identifier=identifier,
            provider=self,
        )

    def add_group(
        self,
        *,
        target_node=None,
        add_action=AddAction.ADD_TO_HEAD,
        name: Optional[str] = None,
    ) -> GroupProxy:
        if not self.moment:
            raise ValueError("No current moment")
        target_node = self._resolve_target_node(target_node)
        identifier = self.server.node_id_allocator.allocate_node_id(1)
        proxy = GroupProxy(identifier=identifier, provider=self)
        self.moment.node_additions.append((proxy, add_action, target_node))
        if name:
            self._annotation_map[identifier] = name
        return proxy

    def add_synth(
        self,
        *,
        synthdef: SynthDef = None,
        target_node=None,
        add_action=AddAction.ADD_TO_HEAD,
        name: Optional[str] = None,
        **settings,
    ) -> SynthProxy:
        if not self.moment:
            raise ValueError("No current moment")
        target_node = self._resolve_target_node(target_node)
        identifier = self.server.node_id_allocator.allocate_node_id(1)
        proxy = SynthProxy(
            identifier=identifier,
            provider=self,
            synthdef=synthdef or default,
            settings=settings,
        )
        self.moment.node_additions.append((proxy, add_action, target_node))
        if name:
            self._annotation_map[identifier] = name
        return proxy

    def boot(self, **kwargs):
        self.server.boot(**kwargs)

    def dispose(self, node_proxy: NodeProxy):
        if not self.moment:
            raise ValueError("No current moment")
        return  # This is currently a no-op

    def free_buffer(self, buffer_: BufferProxy):
        if not self.moment:
            raise ValueError("No current moment")
        self.moment.buffer_removals.append(buffer_)

    def free_bus(self, bus_proxy: BusProxy):
        if not self.moment:
            raise ValueError("No current moment")
        allocator = realtime.Bus._get_allocator(
            bus_proxy.calculation_rate, server=self.server
        )
        allocator.free(cast(int, bus_proxy.identifier))

    def free_bus_group(self, bus_group_proxy: BusGroupProxy):
        if not self.moment:
            raise ValueError("No current moment")
        allocator = realtime.Bus._get_allocator(
            bus_group_proxy.calculation_rate, server=self.server
        )
        allocator.free(cast(int, bus_group_proxy.identifier))

    def free_node(self, node_proxy: NodeProxy):
        if not self.moment:
            raise ValueError("No current moment")
        self.moment.node_removals.append(node_proxy)
        self._annotation_map.pop(node_proxy.identifier, None)

    def move_node(
        self, node_proxy: NodeProxy, add_action: AddAction, target_node: NodeProxy
    ):
        if not self.moment:
            raise ValueError("No current moment")
        target_node = self._resolve_target_node(target_node)
        self.moment.node_reorderings.append((node_proxy, add_action, target_node))

    def quit(self):
        self.server.quit()

    def set_bus(self, bus_proxy: BusProxy, value: float):
        if not self.moment:
            raise ValueError("No current moment")
        elif bus_proxy.calculation_rate != CalculationRate.CONTROL:
            raise ValueError("Can only set control-rate buses")
        self.moment.bus_settings.append((bus_proxy, value))

    def set_node(self, node_proxy: NodeProxy, **settings):
        if not self.moment:
            raise ValueError("No current moment")
        self.moment.node_settings.append((node_proxy, settings))

    def register_osc_callback(
        self, pattern: Tuple[Union[str, float], ...], procedure: Callable
    ) -> OscCallbackProxy:
        identifier = self.server.osc_protocol.register(
            pattern=pattern, procedure=procedure
        )
        return OscCallbackProxy(provider=self, identifier=identifier)

    def unregister_osc_callback(self, proxy: OscCallbackProxy):
        self.server.osc_protocol.unregister(proxy.identifier)



from typing import Optional, Tuple

import supriya.exceptions
from supriya import CalculationRate
from supriya.system import SupriyaValueObject

from ..typing import CalculationRateLike
from .bases import ServerObject

# TODO: Reimplement Bus/BusGroup to stress "leasing" model


class Bus(ServerObject):
    """
    A bus.

    ::

        >>> server = supriya.Server().boot()
        >>> bus = server.add_bus()
        >>> bus
        <+ Bus: 0 (control)>

    ::

        >>> bus.get()
        0.0

    ::

        >>> bus.set(0.5)
        >>> bus.get()
        0.5

    ::

        >>> print(bus)
        c0

    ::

        >>> bus.free()
        <- Bus: ??? (control)>

    """

    ### INITIALIZER ###

    def __init__(
        self,
        bus_group_or_index: Optional[int] = None,
        calculation_rate: CalculationRateLike = CalculationRate.CONTROL,
    ):
        ServerObject.__init__(self)
        bus_group = None
        bus_id = None
        self._bus_id_was_set_manually = False
        if bus_group_or_index is not None:
            self._bus_id_was_set_manually = True
            if isinstance(bus_group_or_index, BusGroup):
                bus_group = bus_group_or_index
            elif isinstance(bus_group_or_index, int):
                bus_id = int(bus_group_or_index)
        self._bus_group = bus_group
        self._bus_id = bus_id
        if calculation_rate is None:
            calculation_rate = "control"
        self._calculation_rate = CalculationRate.from_expr(calculation_rate)
        if self._calculation_rate not in (
            CalculationRate.AUDIO,
            CalculationRate.CONTROL,
        ):
            raise ValueError(calculation_rate)

    ### SPECIAL METHODS ###

    def __float__(self) -> float:
        if self.bus_id is None:
            raise supriya.exceptions.BusNotAllocated
        return float(self.bus_id)

    def __int__(self) -> int:
        if self.bus_id is None:
            raise supriya.exceptions.BusNotAllocated
        return int(self.bus_id)

    def __repr__(self):
        bus_id = self.bus_id
        if bus_id is None:
            bus_id = "???"
        return "<{} {}: {} ({})>".format(
            "+" if self.is_allocated else "-",
            type(self).__name__,
            bus_id,
            self.calculation_rate.name.lower(),
        )

    def __str__(self) -> str:
        """
        Gets map symbol representation of bus.

        ::

            >>> server = supriya.Server().boot()
            >>> control_bus = server.add_bus("control")
            >>> audio_bus = server.add_bus("audio")

        ::

            >>> print(str(control_bus))
            c0

        ::

            >>> print(str(audio_bus))
            a16

        ::

            >>> print(str(control_bus.free()))
            Traceback (most recent call last):
            ...
            supriya.exceptions.BusNotAllocated

        """
        return self.map_symbol

    ### PRIVATE METHODS ###

    @staticmethod
    def _get_allocator(calculation_rate=None, server=None):
        if calculation_rate == CalculationRate.AUDIO:
            allocator = server.audio_bus_allocator
        else:
            allocator = server.control_bus_allocator
        return allocator

    ### PUBLIC METHODS ###

    def allocate(self, server: "supriya.realtime.servers.Server") -> "Bus":
        if self.bus_group is not None:
            return self
        if self.is_allocated:
            raise supriya.exceptions.BusAlreadyAllocated
        ServerObject.allocate(self, server=server)
        if self.bus_id is None:
            allocator = self._get_allocator(
                calculation_rate=self.calculation_rate, server=self.server
            )
            bus_id = allocator.allocate(1)
            if bus_id is None:
                ServerObject.free(self)
                raise ValueError
            self._bus_id = bus_id
        return self

    def free(self) -> "Bus":
        if not self.is_allocated:
            raise supriya.exceptions.BusNotAllocated
        if not self._bus_id_was_set_manually:
            allocator = self._get_allocator(
                calculation_rate=self.calculation_rate, server=self.server
            )
            allocator.free(self.bus_id)
        self._bus_id = None
        ServerObject.free(self)
        return self

    def get(self, completion_callback=None) -> float:
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BusNotAllocated
        elif not self.calculation_rate == CalculationRate.CONTROL:
            raise supriya.exceptions.IncompatibleRate
        request = supriya.commands.ControlBusGetRequest(indices=(self,))
        if callable(completion_callback):
            raise NotImplementedError
        response = request.communicate(server=self.server)
        assert len(response) == 1
        value = response[0].bus_value
        return value

    def set(self, value: float) -> None:
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BusNotAllocated
        elif not self.calculation_rate == CalculationRate.CONTROL:
            raise supriya.exceptions.IncompatibleRate
        request = supriya.commands.ControlBusSetRequest(
            index_value_pairs=((self, value),)
        )
        request.communicate(server=self.server, sync=False)

    ### PUBLIC PROPERTIES ###

    @property
    def bus_group(self) -> Optional["BusGroup"]:
        return self._bus_group

    @property
    def bus_id(self) -> Optional[int]:
        if self._bus_group is not None:
            if self._bus_group.bus_id is not None:
                group_id = self._bus_group.bus_id
                index = self._bus_group.index(self)
                bus_id = group_id + index
                return bus_id
        return self._bus_id

    @property
    def calculation_rate(self) -> CalculationRate:
        return self._calculation_rate

    @property
    def is_allocated(self) -> bool:
        if self.bus_group is not None:
            return self.bus_group.is_allocated
        return self.server is not None

    @property
    def map_symbol(self) -> str:
        if self.bus_id is None:
            raise supriya.exceptions.BusNotAllocated
        if self.calculation_rate == CalculationRate.AUDIO:
            map_symbol = "a"
        else:
            map_symbol = "c"
        map_symbol += str(self.bus_id)
        return map_symbol

    @property
    def server(self) -> Optional["supriya.realtime.servers.Server"]:
        if self.bus_group is not None:
            return self.bus_group.server
        return self._server

    @property
    def value(self) -> Optional[float]:
        if self.server is not None and self.calculation_rate == CalculationRate.CONTROL:
            return self.server._get_control_bus_proxy(self.bus_id).value
        return None


class BusGroup(ServerObject):
    """
    A bus group.

    ::

        >>> server = supriya.Server().boot()
        >>> bus_group = supriya.BusGroup(bus_count=4)
        >>> bus_group
        <- BusGroup{4}: ??? (control)>

    ::

        >>> bus_group.allocate(server)
        <+ BusGroup{4}: 0 (control)>

    ::

        >>> bus_group[2]
        <+ Bus: 2 (control)>

    ::

        >>> for i in range(len(bus_group)):
        ...     bus = bus_group[i]
        ...     value = (i * 0.2) + 0.1
        ...     bus.set(value)
        ...
        >>> bus_values = bus_group.get()

    Values in ``scsynth`` don't necessarily have the same precision as in
    Python, so we'll round them here for display purposes:

    ::

        >>> print([round(value, 1) for value in bus_values])
        [0.1, 0.3, 0.5, 0.7]

    ::

        >>> print(bus_group)
        c0

    ::

        >>> bus_group.free()
        <- BusGroup{4}: ??? (control)>

    """

    ### INITIALIZER ###

    def __init__(
        self, bus_count=1, calculation_rate=CalculationRate.CONTROL, *, bus_id=None
    ):
        ServerObject.__init__(self)
        self._calculation_rate = CalculationRate.from_expr(calculation_rate)
        if self._calculation_rate not in (
            CalculationRate.AUDIO,
            CalculationRate.CONTROL,
        ):
            raise ValueError(self._calculation_rate)
        bus_count = int(bus_count)
        if bus_count <= 0:
            raise ValueError(bus_count)
        self._buses = tuple(
            Bus(bus_group_or_index=self, calculation_rate=self.calculation_rate)
            for _ in range(bus_count)
        )
        if not isinstance(bus_id, (type(None), int)):
            raise ValueError(bus_id)
        self._bus_id = bus_id

    ### SPECIAL METHODS ###

    def __contains__(self, item) -> bool:
        """
        Test if a bus belongs to the bus group.

        ::

            >>> bus_group = supriya.BusGroup(4, "control")
            >>> bus_group[0] in bus_group
            True

        ::

            >>> bus = supriya.Bus("audio")
            >>> bus in bus_group
            False

        """
        # TODO: Should this handle allocated buses that match by ID?
        return self.buses.__contains__(item)

    def __float__(self) -> float:
        if self.bus_id is None:
            raise supriya.exceptions.BusNotAllocated
        return float(self.bus_id)

    def __getitem__(self, item: int) -> Bus:
        """
        Get ``item`` in bus group.

        ::

            >>> server = supriya.Server().boot()
            >>> bus_group = supriya.BusGroup(4, "control").allocate(server)
            >>> bus_group[0]
            <+ Bus: 0 (control)>

        ::

            >>> bus_group[1:]
            <+ BusGroup{3}: 1 (control)>

        """
        if isinstance(item, int):
            return self._buses[item]
        elif isinstance(item, slice):
            indices = item.indices(len(self))
            bus_count = indices[1] - indices[0]
            bus_group = type(self)(
                bus_count=bus_count,
                bus_id=indices[0],
                calculation_rate=self.calculation_rate,
            )
            if self.is_allocated:
                bus_group._server = self.server
            return bus_group

    def __int__(self) -> int:
        if self.bus_id is None:
            raise supriya.exceptions.BusNotAllocated
        return int(self.bus_id)

    def __iter__(self):
        return iter(self.buses)

    def __len__(self) -> int:
        return len(self._buses)

    def __repr__(self) -> str:
        return "<{} {}{{{}}}: {} ({})>".format(
            "+" if self.is_allocated else "-",
            type(self).__name__,
            len(self),
            self.bus_id if self.bus_id is not None else "???",
            self.calculation_rate.name.lower(),
        )

    def __str__(self) -> str:
        """
        Gets map symbol representation of bus group.

        ::

            >>> server = supriya.Server().boot()
            >>> control_bus_group = server.add_bus_group(4, "control")
            >>> audio_bus_group = server.add_bus_group(4, "audio")

        ::

            >>> print(str(control_bus_group))
            c0

        ::

            >>> print(str(audio_bus_group))
            a16

        """
        return self.map_symbol

    ### PUBLIC METHODS ###

    def allocate(self, server: "supriya.realtime.servers.Server") -> "BusGroup":
        if self.is_allocated:
            raise supriya.exceptions.BusAlreadyAllocated
        ServerObject.allocate(self, server=server)
        allocator = Bus._get_allocator(
            calculation_rate=self.calculation_rate, server=self.server
        )
        bus_id = allocator.allocate(len(self))
        if bus_id is None:
            ServerObject.free(self)
            raise ValueError
        self._bus_id = bus_id
        return self

    def fill(self, value: float):
        """
        Fill buses in bus group with ``value``.

        ::

            >>> server = supriya.Server().boot()
            >>> bus_group = server.add_bus_group(4, "control")
            >>> bus_group.get()
            (0.0, 0.0, 0.0, 0.0)

        ::

            >>> bus_group.fill(0.5)

        ::

            >>> bus_group.get()
            (0.5, 0.5, 0.5, 0.5)

        ::

            >>> bus_group = supriya.BusGroup(4, "audio")
            >>> bus_group.fill(0.5)
            Traceback (most recent call last):
            ...
            supriya.exceptions.BusNotAllocated

        ::

            >>> bus_group.allocate(server).fill(0.5)
            Traceback (most recent call last):
            ...
            supriya.exceptions.IncompatibleRate

        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BusNotAllocated
        if self.calculation_rate != CalculationRate.CONTROL:
            raise supriya.exceptions.IncompatibleRate
        index_count_value_triples = [(self.bus_id, len(self), value)]
        request = supriya.commands.ControlBusFillRequest(
            index_count_value_triples=index_count_value_triples
        )
        request.communicate(server=self.server, sync=False)

    def free(self) -> "BusGroup":
        if not self.is_allocated:
            return self
        allocator = Bus._get_allocator(
            calculation_rate=self.calculation_rate, server=self.server
        )
        if allocator:
            allocator.free(self.bus_id)
        self._bus_id = None
        ServerObject.free(self)
        return self

    def get(self) -> Tuple[float, ...]:
        """
        Get bus group values.

        ::

            >>> server = supriya.Server().boot()
            >>> bus_group = supriya.BusGroup(4, "control").allocate(server)
            >>> bus_group.get()
            (0.0, 0.0, 0.0, 0.0)

        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BusNotAllocated
        if self.calculation_rate != CalculationRate.CONTROL:
            raise supriya.exceptions.IncompatibleRate
        index_count_pairs = [(self.bus_id, len(self))]
        request = supriya.commands.ControlBusGetContiguousRequest(
            index_count_pairs=index_count_pairs
        )
        response = request.communicate(server=self.server)
        assert len(response) == 1
        value = response[0].bus_values
        return value

    def index(self, item) -> int:
        return self.buses.index(item)

    def set(self, *values: float) -> None:
        """
        Set bus group values.

        ::

            >>> server = supriya.Server().boot()
            >>> bus_group = supriya.BusGroup(4, "control").allocate(server)
            >>> bus_group.get()
            (0.0, 0.0, 0.0, 0.0)

        ::

            >>> bus_group.set(-0.5, 0.5, -0.5, 0.5)
            >>> bus_group.get()
            (-0.5, 0.5, -0.5, 0.5)

        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BusNotAllocated(self)
        if self.calculation_rate != CalculationRate.CONTROL:
            raise supriya.exceptions.IncompatibleRate(self)
        if len(values) != len(self):
            raise ValueError(values)
        request = supriya.commands.ControlBusSetContiguousRequest(
            index_values_pairs=[(self, values)]
        )
        request.communicate(self.server, sync=False)

    ### PUBLIC PROPERTIES ###

    @property
    def bus_id(self) -> Optional[int]:
        return self._bus_id

    @property
    def buses(self) -> Tuple[Bus, ...]:
        return self._buses

    @property
    def calculation_rate(self) -> CalculationRate:
        return self._calculation_rate

    @property
    def is_allocated(self) -> bool:
        return self.server is not None

    @property
    def map_symbol(self) -> str:
        if self.bus_id is None:
            raise supriya.exceptions.BusNotAllocated
        if self.calculation_rate == CalculationRate.AUDIO:
            map_symbol = "a"
        else:
            map_symbol = "c"
        map_symbol += str(self.bus_id)
        return map_symbol


class BusProxy(SupriyaValueObject):
    """
    A buffer proxy.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_bus_id", "_calculation_rate", "_server", "_value")

    ### INITIALIZER ###

    def __init__(
        self,
        *,
        bus_id: int,
        calculation_rate: CalculationRateLike,
        server: "supriya.realtime.servers.Server",
        value: float = 0.0,
    ):
        self._bus_id = bus_id
        self._calculation_rate = CalculationRate.from_expr(calculation_rate)
        self._server = server
        self._value = value

    ### SPECIAL METHODS ###

    def __float__(self) -> float:
        return float(self.bus_id)

    def __int__(self) -> int:
        return int(self.bus_id)

    def __str__(self) -> str:
        return self.map_symbol

    ### PUBLIC PROPERTIES ###

    @property
    def bus_id(self) -> int:
        return self._bus_id

    @property
    def calculation_rate(self) -> CalculationRate:
        return self._calculation_rate

    @property
    def map_symbol(self) -> str:
        if self.calculation_rate == CalculationRate.AUDIO:
            map_symbol = "a"
        else:
            map_symbol = "c"
        map_symbol += str(self.bus_id)
        return map_symbol

    @property
    def server(self) -> "supriya.realtime.servers.Server":
        return self._server

    @property
    def value(self) -> float:
        return self._value


class AudioInputBusGroup(BusGroup):
    """
    Audio input bus group.

    Allocated automatically on server boot.

    ::

        >>> server = supriya.Server().boot()
        >>> bus_group = server.audio_input_bus_group
        >>> bus_group
        <+ AudioInputBusGroup{8}: 8 (audio)>

    ::

        >>> bus_group.is_allocated
        True

    """

    ### INITIALIZER ###

    def __init__(self, server):
        import supriya.realtime
        import supriya.synthdefs

        assert isinstance(server, supriya.realtime.Server)
        assert server.is_running
        BusGroup.__init__(
            self,
            bus_count=server.options.input_bus_channel_count,
            calculation_rate=supriya.CalculationRate.AUDIO,
        )
        self._bus_id = server.options.output_bus_channel_count
        self._server = server

    ### PUBLIC METHODS ###

    def allocate(self):
        pass

    def free(self):
        pass


class AudioOutputBusGroup(BusGroup):
    """
    Audio output bus group.

    Allocated automatically on server boot.

    ::

        >>> server = supriya.Server().boot()
        >>> bus_group = server.audio_output_bus_group
        >>> bus_group
        <+ AudioOutputBusGroup{8}: 0 (audio)>

    ::

        >>> bus_group.is_allocated
        True

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, server):
        import supriya.realtime
        import supriya.synthdefs

        assert isinstance(server, supriya.realtime.Server)
        assert server.is_running
        BusGroup.__init__(
            self,
            bus_count=server.options.input_bus_channel_count,
            calculation_rate=supriya.CalculationRate.AUDIO,
        )
        self._bus_id = 0
        self._server = server

    ### PUBLIC METHODS ###

    def allocate(self):
        pass

    def free(self):
        pass



from typing import NamedTuple, Union

from .messages import OscBundle, OscMessage


class CaptureEntry(NamedTuple):
    timestamp: float
    label: str
    message: Union[OscMessage, OscBundle]


class Capture:

    ### INITIALIZER ###

    def __init__(self, osc_protocol):
        self.osc_protocol = osc_protocol
        self.messages = []

    ### SPECIAL METHODS ###

    def __enter__(self):
        self.osc_protocol.captures.add(self)
        self.messages[:] = []
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.osc_protocol.captures.remove(self)

    def __iter__(self):
        return iter(self.messages)

    def __len__(self):
        return len(self.messages)

    ### PUBLIC PROPERTIES ###

    @property
    def received_messages(self):
        return [
            (timestamp, osc_message)
            for timestamp, label, osc_message in self.messages
            if label == "R"
        ]

    @property
    def sent_messages(self):
        return [
            (timestamp, osc_message)
            for timestamp, label, osc_message in self.messages
            if label == "S"
        ]



import dataclasses
import enum
from typing import Callable, Dict, NamedTuple, Optional, Tuple, Union


class EventType(enum.IntEnum):
    CHANGE = 0
    SCHEDULE = 1


class TimeUnit(enum.IntEnum):
    BEATS = 0
    SECONDS = 1
    MEASURES = 2


class ClockState(NamedTuple):
    beats_per_minute: float
    initial_seconds: float
    previous_measure: int
    previous_offset: float
    previous_seconds: float
    previous_time_signature_change_offset: float
    time_signature: Tuple[int, int]


@dataclasses.dataclass(frozen=True)
class Moment:
    __slots__ = (
        "beats_per_minute",
        "measure",
        "measure_offset",
        "offset",
        "seconds",
        "time_signature",
    )
    beats_per_minute: float
    measure: int
    measure_offset: float
    offset: float
    seconds: float
    time_signature: Tuple[int, int]


class CallbackCommand(NamedTuple):
    args: Optional[Tuple]
    event_id: int
    event_type: int
    kwargs: Optional[Dict]
    procedure: Callable
    quantization: Optional[str]
    schedule_at: float
    time_unit: Optional[int]


class CallbackEvent(NamedTuple):
    seconds: float
    event_type: int
    event_id: int
    measure: Optional[int]
    offset: Optional[float]
    procedure: Callable
    args: Optional[Tuple]
    kwargs: Optional[Dict]
    invocations: int

    def __hash__(self):
        return hash((type(self), self.event_id))


class ChangeCommand(NamedTuple):
    beats_per_minute: Optional[float]
    event_id: int
    event_type: int
    quantization: Optional[str]
    schedule_at: float
    time_signature: Optional[Tuple[int, int]]
    time_unit: Optional[int]


class ChangeEvent(NamedTuple):
    seconds: float
    event_type: int
    event_id: int
    measure: Optional[int]
    offset: Optional[float]
    beats_per_minute: float
    time_signature: Tuple[int, int]

    def __hash__(self):
        return hash((type(self), self.event_id))


class ClockContext(NamedTuple):
    current_moment: Moment
    desired_moment: Moment
    event: Union[CallbackEvent, ChangeEvent]



import bisect

import supriya.synthdefs
from supriya.nonrealtime.bases import SessionObject


class Bus(SessionObject):
    """
    A non-realtime bus.

    ::

        >>> import supriya.nonrealtime
        >>> session = supriya.nonrealtime.Session()
        >>> bus = session.add_bus("control")
        >>> print(repr(bus))
        <Bus(<Session>, calculation_rate=CalculationRate.CONTROL, session_id=0)>

    ::

        >>> with session.at(1):
        ...     bus.set_(0.5)
        ...
        >>> with session.at(3):
        ...     bus.set_(0.75)
        ...

    ::

        >>> for offset in range(5):
        ...     with session.at(offset):
        ...         value = bus.get()
        ...         print(offset, value)
        ...
        0 0.0
        1 0.5
        2 0.5
        3 0.75
        4 0.75

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_bus_group",
        "_calculation_rate",
        "_events",
        "_session",
        "_session_id",
    )

    ### INITIALIZER ###

    def __init__(self, session, bus_group=None, calculation_rate=None, session_id=None):
        import supriya.nonrealtime

        SessionObject.__init__(self, session)
        self._session_id = session_id
        if bus_group is not None:
            assert isinstance(bus_group, supriya.nonrealtime.BusGroup)
        self._bus_group = bus_group
        assert calculation_rate is not None
        calculation_rate = supriya.CalculationRate.from_expr(calculation_rate)
        self._calculation_rate = calculation_rate
        self._events = []

    ### SPECIAL METHODS ###

    def __repr__(self):
        return "<{}>".format(super(Bus, self).__repr__())

    def __float__(self):
        return float(self._session_id)

    def __int__(self):
        return int(self._session_id)

    def __str__(self):
        map_symbol = "c"
        if self.calculation_rate == supriya.CalculationRate.AUDIO:
            map_symbol = "a"
        session_id = self._session_id
        if session_id is None:
            session_id = "?"
        elif isinstance(session_id, tuple):
            session_id = "{}:{}".format(session_id[0], session_id[1])
        return "{map_symbol}{session_id}".format(
            map_symbol=map_symbol, session_id=session_id
        )

    ### PRIVATE METHODS ###

    def _get_at_offset(self, offset):
        events = self._events
        if not events:
            return 0.0
        index = bisect.bisect_left(events, (offset, 0.0))
        if len(events) <= index:
            old_offset, value = events[-1]
        else:
            old_offset, value = events[index]
        if old_offset == offset:
            return value
        index -= 1
        if index < 0:
            return 0.0
        _, value = events[index]
        return value

    def _set_at_offset(self, offset, value):
        assert self.calculation_rate == supriya.CalculationRate.CONTROL
        events = self._events
        event = (offset, value)
        if not events:
            events.append(event)
            return
        index = bisect.bisect_left(events, event)
        if len(events) <= index:
            events.append(event)
        old_offset, _ = events[index]
        if old_offset == offset:
            events[index] = event
        else:
            events.insert(index, event)

    ### PUBLIC METHODS ###

    @SessionObject.require_offset
    def get(self, offset=None):
        value = self._get_at_offset(offset)
        return value

    def get_map_symbol(self, bus_id):
        import supriya.synthdefs

        if self.calculation_rate == supriya.CalculationRate.AUDIO:
            map_symbol = "a"
        else:
            map_symbol = "c"
        map_symbol += str(bus_id)
        return map_symbol

    @SessionObject.require_offset
    def set_(self, value, offset=None):
        self._set_at_offset(offset, value)

    ### PUBLIC PROPERTIES ###

    @property
    def bus_group(self):
        return self._bus_group

    @property
    def calculation_rate(self):
        return self._calculation_rate

    @property
    def session_id(self):
        return self._session_id

    @property
    def start_offset(self):
        return float("-inf")

    @property
    def stop_offset(self):
        return float("inf")


class BusGroup(SessionObject):
    """
    A non-realtime bus group.

    ::

        >>> import supriya.nonrealtime
        >>> session = supriya.nonrealtime.Session()
        >>> bus_group = session.add_bus_group(3)
        >>> print(repr(bus_group))
        <BusGroup(<Session>, bus_count=3, calculation_rate=CalculationRate.CONTROL, session_id=0)>

    ::

        >>> with session.at(1):
        ...     bus_group[0].set_(0.333)
        ...
        >>> with session.at(3):
        ...     bus_group.fill(0.5)
        ...
        >>> with session.at(5):
        ...     bus_group[1].set_(0.666)
        ...     bus_group[2].set_(0.75)
        ...

    ::

        >>> for offset in range(7):
        ...     with session.at(offset):
        ...         values = bus_group.get()
        ...         print(offset, values)
        ...
        0 [0.0, 0.0, 0.0]
        1 [0.333, 0.0, 0.0]
        2 [0.333, 0.0, 0.0]
        3 [0.5, 0.5, 0.5]
        4 [0.5, 0.5, 0.5]
        5 [0.5, 0.666, 0.75]
        6 [0.5, 0.666, 0.75]

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_buses", "_calculation_rate", "_session", "_session_id")

    ### INITIALIZER ###

    def __init__(self, session, bus_count=1, calculation_rate=None, session_id=None):
        import supriya.nonrealtime

        SessionObject.__init__(self, session)
        self._session_id = session_id
        assert calculation_rate is not None
        calculation_rate = supriya.CalculationRate.from_expr(calculation_rate)
        self._calculation_rate = calculation_rate
        bus_count = int(bus_count)
        assert 0 < bus_count
        self._buses = tuple(
            supriya.nonrealtime.Bus(
                session,
                bus_group=self,
                calculation_rate=self.calculation_rate,
                session_id=(session_id, i),
            )
            for i in range(bus_count)
        )

    ### SPECIAL METHODS ###

    def __contains__(self, item):
        return self.buses.__contains__(item)

    def __float__(self):
        return float(self._session_id)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._buses[item]
        elif isinstance(item, slice):
            return tuple(self._buses[item])

    def __int__(self):
        return int(self._session_id)

    def __iter__(self):
        return iter(self.buses)

    def __len__(self):
        return len(self._buses)

    def __repr__(self):
        return "<{}>".format(super(BusGroup, self).__repr__())

    def __str__(self):
        map_symbol = "c"
        if self.calculation_rate == supriya.CalculationRate.AUDIO:
            map_symbol = "a"
        session_id = self._session_id
        if session_id is None:
            session_id = "?"
        return "{map_symbol}{session_id}".format(
            map_symbol=map_symbol, session_id=session_id
        )

    ### PUBLIC METHODS ###

    def fill(self, value):
        assert self.session._active_moments
        offset = self.session._active_moments[-1].offset
        for bus in self:
            bus._set_at_offset(offset, value)

    def get(self):
        assert self.session._active_moments
        offset = self.session._active_moments[-1].offset
        values = [bus._get_at_offset(offset) for bus in self]
        return values

    def get_map_symbol(self, bus_id):
        import supriya.synthdefs

        if self.calculation_rate == supriya.CalculationRate.AUDIO:
            map_symbol = "a"
        else:
            map_symbol = "c"
        map_symbol += str(bus_id)
        return map_symbol

    def index(self, item):
        return self.buses.index(item)

    ### PUBLIC PROPERTIES ###

    @property
    def bus_count(self):
        return len(self._buses)

    @property
    def buses(self):
        return self._buses

    @property
    def calculation_rate(self):
        return self._calculation_rate

    @property
    def session_id(self):
        return self._session_id


class AudioInputBusGroup(BusGroup):
    """
    A non-realtime audio input bus group.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, session):
        calculation_rate = supriya.CalculationRate.AUDIO
        bus_count = session.options.input_bus_channel_count
        BusGroup.__init__(
            self, session, bus_count=bus_count, calculation_rate=calculation_rate
        )


class AudioOutputBusGroup(BusGroup):
    """
    A non-realtime audio output bus group.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, session):
        calculation_rate = supriya.CalculationRate.AUDIO
        bus_count = session.options.output_bus_channel_count
        BusGroup.__init__(
            self, session, bus_count=bus_count, calculation_rate=calculation_rate
        )
